# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Import for autodoc -----------------------------------------------------

import os
import sys

sys.path.insert(0, os.path.abspath("../src/idinn"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "idinn"
copyright = "2024, J. Li, T. Asikis, I. Fragkos, L. Böttcher"
author = "J. Li, T. Asikis, I. Fragkos, L. Böttcher"
release = "0.2.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "sphinxcontrib.youtube",
]
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



import inspect
import logging
import os
import queue
import shutil
import threading
import time
from logging import LogRecord
from logging.handlers import QueueHandler
from os import write

import numpy as np
import pandas as pd
import plotly.express as px
import streamlit as st
import torch
from torch.utils.tensorboard import SummaryWriter
from torchview import draw_graph

from idinn.dual_controller import DualSourcingNeuralController
from idinn.demand import CustomDemand, UniformDemand
from idinn.sourcing_model import DualSourcingModel

from utils import tflog2pandas
from io import StringIO

st.set_page_config(layout="wide")

if 'log_queue' not in st.session_state:
    st.session_state['log_queue'] = queue.Queue()

queue_handler = QueueHandler(st.session_state['log_queue'])
formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
queue_handler.setFormatter(formatter)

root_logger = logging.getLogger()
root_logger.setLevel(logging.DEBUG)
root_logger.addHandler(queue_handler)

# Color mapping for log levels
LOG_COLORS = {
    'DEBUG': 'blue',
    'INFO': 'green',
    'WARNING': 'orange',
    'ERROR': 'red',
    'CRITICAL': 'purple'
}


@st.fragment(run_every="1s")
def write_logs():
    process_log_queue()
    messages = list(map(lambda log: log.msg,st.session_state.logs[-100:]))
    prev_log = None
    for log in messages:  # Show last 100 logs
        try:
            parts = log.split(' - ')
            if len(parts) >= 3:
                timestamp, level, message = parts[0], parts[1], ' - '.join(parts[2:])
                color = LOG_COLORS.get(level, 'gray')
                if log != prev_log:
                    st.markdown(f"`{timestamp}` :{color}[**{level}**] {message}")
                prev_log = log
        except Exception as e:
            st.write(f"Error parsing log: {str(e)}")
            time.sleep(0.1)

if torch.cuda.is_available():
    torch.set_default_device('cuda')
    torch.classes.__path__ = [os.path.join(torch.__path__[0], torch.classes.__file__)]

    def update_log_display(log_placeholder):
        """Render logs with color coding in markdown"""
        with log_placeholder.container():
           write_logs()

    def process_log_queue():
        """Check queue for new logs and update session state"""
        while True:
            try:
                log = st.session_state['log_queue'].get_nowait()
                st.session_state.logs.append(log)
            except queue.Empty:
                break

# Session State Initialization
for key in ["training", "trainingnow", "demand_generator", "dual_sourcing_model", "demand_controller"]:
    if key not in st.session_state:
        st.session_state[key] = 0 if key == "training" else False if key == "trainingnow" else None
if 'logs' not in st.session_state:
    st.session_state['logs'] = []

# Page Header
st.header("Inventory Dynamics–Informed Neural Networks")
st.markdown(
    "Welcome to Inventory Dynamics–Informed Neural Networks! This application generates ordering policies from expedited and regular suppliers. Use the sidebar on the left to select a demand model, choose your preferred solver, and view the results after fitting."
)

# Tabs
tab1, tab2, tab3, tab4 = st.tabs(["Demand Generation", "Dual Sourcing Model", "NN Controller", "Result"])

# Demand Generation Tab
with tab1:
    c1, c2 = st.columns([1, 3])

    with c1:
        submitted = submitted2 = False
        st.subheader("Demand Generation")

        demand_type = st.radio(label="Please choose demand type", options=["Uniform", "File"])
        if demand_type == "Uniform":
            with st.form(key="uniform_demand"):
                low = st.number_input("Minimum demand", value=1, step=1, format="%i", max_value=(1 << 53) - 1)
                high = st.number_input("Maximum demand", value=4, step=1, format="%i", min_value=0)
                cg1, cg2 = st.columns([1, 2])
                with cg1:
                    pass
                with cg2:
                    submitted = st.form_submit_button("Generate")

                if submitted and high >= low:
                    st.success(
                        f"Successfully generated uniform demand within range: [{np.floor(low)}, {np.floor(high)}]!")
                    st.session_state["training"] = 0
                    st.session_state["demand_generator"] = UniformDemand(low, high)
                elif submitted:
                    st.error("Please resubmit and make sure that maximum demand is greater or equal to minimum demand.")
                    st.session_state["training"] = 0

        elif demand_type == "File":
            with st.form(key="uniform_demand"):
                uploaded_file = st.file_uploader(
                    label="Please upload a single column file with demand values. Each row represents a timestep and each element represents a demand value.")
                cgg1, cgg2 = st.columns([1, 2])
                with cgg1:
                    submitted2 = st.form_submit_button("Generate")
                with cgg2:
                    if submitted2:
                        try:
                            df = pd.read_csv(uploaded_file)
                            st.success(f"Successfully uploaded file and contains {df.shape[0]} demand points!")
                            st.session_state["demand_generator"] = CustomDemand(torch.tensor(df.iloc[:, 0].values))
                            st.session_state["training"] = 0
                        except Exception:
                            st.error("Could not parse file! Please try again!")

    with c2:
        if st.session_state["demand_generator"] is not None:
            all_demands = [st.session_state["demand_generator"].sample(1).item() for _ in range(100)]
            c2c1, c2c2 = st.columns(2)
            with c2c1:
                fig = px.line(y=all_demands, line_shape="hv").update_layout(xaxis_title="Periods", yaxis_title="Demand",
                                                                            title="Typical Demand Trajectory")
                st.plotly_chart(fig, use_container_width=True)
            with c2c2:
                fig = px.histogram(x=all_demands).update_layout(xaxis_title="Demand", yaxis_title="Frequency",
                                                                title="Demand Distribution")
                st.plotly_chart(fig, use_container_width=True)

# Dual Sourcing Model Tab
with tab2:
    st.subheader("Dual Sourcing Model")
    with st.form(key="Dual Sourcing Model"):
        c1t2, c2t2 = st.columns(2)
        with c1t2:
            regular_lead_time = np.int32(
                st.number_input("Regular lead time:", value=2, min_value=0, format="%i", step=1))
            expedited_lead_time = np.int32(
                st.number_input("Expedited lead time:", value=0, min_value=0, format="%i", step=1))
            batch_size = np.int32(
                st.number_input("Minibatch size for demand trajectories:", value=16, min_value=0, format="%i", step=1))
            init_inventory = np.int32(st.number_input("Initial inventory:", value=6, min_value=0, format="%i", step=1))

        with c2t2:
            regular_order_cost = np.int32(
                st.number_input("Regular order cost:", value=0, min_value=0, format="%i", step=1))
            expedited_order_cost = np.int32(
                st.number_input("Expedited order cost:", value=20, min_value=0, format="%i", step=1))
            holding_cost = np.int32(st.number_input("Holding cost:", value=5, min_value=0, format="%i", step=1))
            shortage_cost = np.int32(st.number_input("Shortage cost:", value=495, min_value=0, format="%i", step=1))

        model_params = {
            "regular_lead_time": regular_lead_time,
            "expedited_lead_time": expedited_lead_time,
            "regular_order_cost": regular_order_cost,
            "expedited_order_cost": expedited_order_cost,
            "holding_cost": holding_cost,
            "shortage_cost": shortage_cost,
            "batch_size": batch_size,
            "init_inventory": init_inventory,
            "demand_generator": st.session_state["demand_generator"]
        }

        cc1, cc2 = st.columns([1, 4])
        with cc1:
            pressed2 = st.form_submit_button("Create Sourcing Model")
        with cc2:
            if pressed2:
                st.session_state["training"] = 0
                st.session_state["dual_sourcing_model"] = DualSourcingModel(**model_params)
                st.success("Successfully created sourcing model!")

# Controller Definition Tab
with tab3:
    st.subheader("Controller Definition")
    c1, c2 = st.columns(2)
    with c1:
        activation_map = {
            "CELU": torch.nn.CELU,
            "ReLU": torch.nn.ReLU,
            "LeakyReLU": torch.nn.LeakyReLU,
            "ELU": torch.nn.ELU,
            "Tanh": torch.nn.Tanh,
            "Sigmoid": torch.nn.Sigmoid,
            "SiLU": torch.nn.SiLU,
            "GELU": torch.nn.GELU,
            "TahnShrink": torch.nn.Tanhshrink,
        }
        activation_id = st.selectbox("Activation function of hidden layers:", options=list(activation_map.keys()))
        activation_signature = inspect.signature(activation_map[activation_id])
        kwargs = {v.name: st.number_input(f"Value for activation function's parameter: {v.name}",
                                          value=(v.default if v.default is not None else 0.0)) for k, v in
                  activation_signature.parameters.items() if v.annotation in [float, int]}

        layer_sizes = st.text_area(label="Layer sizes:", value="128, 64, 32, 16, 8, 4",
                                   help="A comma separated list of integers, indicating neurons per layer, starting from the first hidden layer (leftmost)")
        sourcing_periods = st.number_input("Number of training sourcing periods:", value=50, min_value=1, format="%i",
                                           step=1)
        validation_sourcing_periods = st.number_input("Number of validation sourcing periods:", value=1000, min_value=1,
                                                      format="%i", step=1)
        epochs = st.number_input("Number of training epochs:", value=2000, min_value=1, format="%i", step=1)
        seed = st.number_input("Seed:", value=1234, min_value=1, format="%i", step=1)
        try:
            layer_sizes = list(map(int, layer_sizes.split(",")))
        except Exception:
            st.error("Provided input cannot be parsed to layers.")

        st.session_state["demand_controller"] = DualSourcingNeuralController(hidden_layers=layer_sizes,
                                                                             activation=activation_map[activation_id](
                                                                                 **kwargs))

    with c2:
        x = torch.linspace(-10, 10, 100)
        fig = px.line(x=x.cpu().numpy(), y=activation_map[activation_id](**kwargs)(x).cpu().numpy(),
                      title=f"Activation shape: {activation_id}")
        st.plotly_chart(fig, use_container_width=True)

        if st.session_state["demand_controller"] and st.session_state["dual_sourcing_model"]:
            st.info(
                "Graph plot for minibatch size of 4 (for illustration purposes). Click top right corner to enlarge!")
            st.session_state["demand_controller"].init_layers(regular_lead_time=regular_lead_time,
                                                              expedited_lead_time=expedited_lead_time)
            pre_input_tensors = torch.rand([4, 1]), torch.rand([4, max(regular_lead_time, 1)]), torch.rand(
                [4, max(regular_lead_time, 1)])

            input_tensor = st.session_state["demand_controller"].prepare_inputs(*pre_input_tensors,
                                                                                sourcing_model=st.session_state[
                                                                                    "dual_sourcing_model"])

            # input_sizes = [torch.Size([4, 1]), torch.Size([4, max(regular_lead_time, 1)]), torch.Size([4, max(regular_lead_time, 1)])]
            input_sizes = [input_tensor.shape]
            model_graph = draw_graph(st.session_state["demand_controller"].model, input_size=input_sizes)
            model_graph.visual_graph.attr("graph", rankdir="LR")
            st.graphviz_chart(model_graph.visual_graph)
        elif not st.session_state["dual_sourcing_model"]:
            st.warning("Please define a dual sourcing model to generate NN architecture graph!")

    log_placeholder = st.empty()

    def click():
        st.session_state["training"] += 1
        if os.path.exists("runs/dual_sourcing_model"):
            shutil.rmtree("runs/dual_sourcing_model")
        st.session_state["trainingnow"] = True

        def fit_model(controller,
                      sourcing_model,
                      sourcing_periods,
                      validation_sourcing_periods,
                      epochs,
                      seed
                      ):
            if torch.cuda.is_available():
                torch.set_default_device('cuda')
                torch.classes.__path__ = [os.path.join(torch.__path__[0], torch.classes.__file__)]

            controller.fit(
                sourcing_model=sourcing_model,
                sourcing_periods=sourcing_periods,
                validation_sourcing_periods=validation_sourcing_periods,
                epochs=epochs,
                tensorboard_writer=SummaryWriter("runs/dual_sourcing_model"),
                seed=seed,
            )
            return controller
        update_log_display(log_placeholder)
        fit_model(*[st.session_state["demand_controller"],
                                                 st.session_state["dual_sourcing_model"],
                                                 sourcing_periods,
                                                 validation_sourcing_periods,
                                                 epochs,
                                                 seed
                                                 ])
        # st.session_state["demand_controller"].fit(
        #     sourcing_model=st.session_state["dual_sourcing_model"],
        #     sourcing_periods=sourcing_periods,
        #     validation_sourcing_periods=validation_sourcing_periods,
        #     epochs=epochs,
        #     tensorboard_writer=SummaryWriter("runs/dual_sourcing_model"),
        #     seed=seed,
        # )
        st.session_state["trainingnow"] = False
        st.success("Training complete!")

    def training_fragment():
        cf1, cf2 = st.columns([1, 4])
        with cf1:
            pressed = st.button("Fit Controller")
        with cf2:
            if pressed:
                click()

    with st.container() as button_cont:
        training_fragment()
    # Display logs in a markdown window

# Results Tab
with tab4:
    if all(st.session_state[key] for key in ["demand_controller", "dual_sourcing_model", "demand_generator"]):
        if os.path.exists("runs/dual_sourcing_model") and st.session_state["training"] > 0:
            try:
                t4c1, t4c2 = st.columns(2)
                with t4c1:
                    tsb_df = tflog2pandas("runs/dual_sourcing_model")
                    print(tsb_df)
                    fig = px.line(tsb_df, x="step", y="value", color="metric", line_shape="hv",
                                  title="Learning Curves").update_layout(yaxis_title="Avg. Cost per Period",
                                                                         xaxis_title="Epoch")
                    st.plotly_chart(fig)
                with t4c2:
                    past_inventories, past_regular_orders, past_expedited_orders = st.session_state[
                        "demand_controller"].simulate(sourcing_model=st.session_state["dual_sourcing_model"],
                                                      sourcing_periods=sourcing_periods)
                    df_past = pd.DataFrame({"Inventory": past_inventories, "Regular Orders": past_regular_orders,
                                            "Expedited Orders": past_expedited_orders})
                    fig = px.line(df_past, line_shape="hv").update_layout(xaxis_title="Periods", yaxis_title="Quantity",
                                                                          title="Sample Optimization Trajectory")
                    st.plotly_chart(fig)
            except Exception as e:
                st.warning("No available model, please make sure you submit the previous steps!")
                st.error(e)
    else:
        if not st.session_state["demand_generator"]:
            st.warning("No demand generator is chosen for the current configuration, please define one.")
        if not st.session_state["dual_sourcing_model"]:
            st.warning("No sourcing model is defined for the current configuration, please define one.")
        if not st.session_state["demand_controller"]:
            st.warning(
                "No model chosen or trained for the current configuration, please go to the previous step and fit a model.")



from abc import ABCMeta, abstractmethod
import torch


class BaseDualController(metaclass=ABCMeta):
    @abstractmethod
    def fit(self, sourcing_model):
        """
        Fit the controller to the sourcing model.
        """
        pass

    @abstractmethod
    def predict(self, current_inventory, past_regular_orders=None, past_expedited_orders=None, output_tensor=False):
        """
        Predict the replenishment order quantity.

        Parameters
        ----------
        current_inventory : int, or torch.Tensor
            Current inventory.
        past_regular_orders : list, or torch.Tensor, optional
            Past regular orders. If the length of `past_regular_orders` is lower than `regular_lead_time`, it will be padded with zeros. If the length of `past_regular_orders` is higher than `regular_lead_time`, only the last `regular_lead_time` orders will be used during inference.
        past_expedited_orders : list, or torch.Tensor, optional
            Past expedited orders. If the length of `past_expedited_orders` is lower than `expedited_lead_time`, it will be padded with zeros. If the length of `past_expedited_orders` is higher than `expedited_lead_time`, only the last `expedited_lead_time` orders will be used during inference.
        output_tensor : bool, default is False
            If True, the replenishment order quantity will be returned as a torch.Tensor. Otherwise, it will be returned as an integer.
        """
        pass

    @abstractmethod
    def reset(self):
        """
        Reset the controller to the initial state.
        """
        pass

    def _check_current_inventory(self, current_inventory):
        """
        Check and convert types of `current_inventory` for `predict()`.
        """
        if isinstance(current_inventory, int):
            current_inventory = torch.tensor([[current_inventory]], dtype=torch.float32)
        elif isinstance(current_inventory, torch.Tensor):
            pass
        else:
            raise TypeError("`current_inventory`'s type is not supported.")

        return current_inventory
    
    def _check_past_orders(self, past_orders, lead_time):
        """
        Check and convert types of `past_regular_orders` for `predict()`.
        Pad `past_orders` with zeros if either is too short.
        """
        if past_orders is None:
            past_orders = torch.zeros(1, lead_time)
        elif isinstance(past_orders, list):
            past_orders = torch.tensor([past_orders], dtype=torch.float32)
        elif isinstance(past_orders, torch.Tensor):
            pass
        else:
            raise TypeError("`past_orders`'s type is not supported.")
        
        order_len = past_orders.shape[1]
        if order_len < lead_time:
            return torch.nn.functional.pad(past_orders, (lead_time - order_len, 0))
        else:
            return past_orders

    def get_last_cost(self, sourcing_model):
        """
        Calculate the cost for the latest period of the sourcing model.

        Parameters
        ----------
        sourcing_model : SourcingModel
            The sourcing model.

        Returns
        -------
        float
            The last cost.
        """
        last_regular_q = sourcing_model.get_last_regular_order()
        last_expedited_q = sourcing_model.get_last_expedited_order()
        regular_order_cost = sourcing_model.get_regular_order_cost()
        expedited_order_cost = sourcing_model.get_expedited_order_cost()
        holding_cost = sourcing_model.get_holding_cost()
        shortage_cost = sourcing_model.get_shortage_cost()
        current_inventory = sourcing_model.get_current_inventory()
        last_cost = (
            regular_order_cost * last_regular_q
            + expedited_order_cost * last_expedited_q
            + holding_cost * torch.relu(current_inventory)
            + shortage_cost * torch.relu(-current_inventory)
        )
        return last_cost

    def get_total_cost(self, sourcing_model, sourcing_periods, seed=None):
        """
        Calculate the total cost for dual-sourcing optimization.

        Parameters
        ----------
        sourcing_model : SourcingModel
            The sourcing model.
        sourcing_periods : int
            Number of sourcing periods.
        seed : int, optional
            Random seed for reproducibility.

        Returns
        -------
        float
            The total cost.
        """
        if seed is not None:
            torch.manual_seed(seed)

        total_cost = 0
        for i in range(sourcing_periods):
            current_inventory = sourcing_model.get_current_inventory()
            past_regular_orders = sourcing_model.get_past_regular_orders()
            past_expedited_orders = sourcing_model.get_past_expedited_orders()
            regular_q, expedited_q = self.predict(
                current_inventory, past_regular_orders, past_expedited_orders, output_tensor=True
            )
            sourcing_model.order(regular_q, expedited_q)
            last_cost = self.get_last_cost(sourcing_model)
            total_cost += last_cost.mean()
        return total_cost

    def get_average_cost(self, sourcing_model, sourcing_periods, seed=None):
        """
        Calculate the average cost for dual-sourcing controllers.

        Parameters
        ----------
        sourcing_model : DualSourcingModel
            The sourcing model.
        sourcing_periods : int
            Number of sourcing periods.
        seed : int, optional
            Random seed for reproducibility.

        Returns
        -------
        float
            The average cost.
        """
        return (
            self.get_total_cost(sourcing_model, sourcing_periods, seed)
            / sourcing_periods
        )

    def simulate(self, sourcing_model, sourcing_periods, seed=None):
        """
        Simulate the sourcing model's output using the given controller.

        Parameters
        ----------
        sourcing_model : DualSourcingModel
            The sourcing model.
        sourcing_periods : int
            Number of sourcing periods.
        seed : int, optional
            Random seed for reproducibility.

        Returns
        -------
        past_inventories : list
            List of past inventories.
        past_regular_orders : list
            List of past regular orders.
        past_expedited_orders : list
            List of past expedited orders.

        """
        if seed is not None:
            torch.manual_seed(seed)
        sourcing_model.reset(batch_size=1)
        for i in range(sourcing_periods):
            current_inventory = sourcing_model.get_current_inventory()
            past_regular_orders = sourcing_model.get_past_regular_orders()
            past_expedited_orders = sourcing_model.get_past_expedited_orders()
            regular_q, expedited_q = self.predict(
                current_inventory, past_regular_orders, past_expedited_orders
            )
            sourcing_model.order(regular_q, expedited_q)
        past_inventories = sourcing_model.get_past_inventories()[0, :].detach().cpu().numpy()
        past_regular_orders = (
            sourcing_model.get_past_regular_orders()[0, :].detach().cpu().numpy()
        )
        past_expedited_orders = (
            sourcing_model.get_past_expedited_orders()[0, :].detach().cpu().numpy()
        )
        return past_inventories, past_regular_orders, past_expedited_orders

    def plot(self, sourcing_model, sourcing_periods, linewidth=1, seed=None):
        """
        Plot the inventory and order quantities.

        Parameters
        ----------
        sourcing_model : DualSourcingModel
            The sourcing model.
        sourcing_periods : int
            Number of sourcing periods.
        linewidth : int, default is 1
            Width of the line in the step plots.
        seed : int, optional
            Random seed for reproducibility.
        """
        from matplotlib import pyplot as plt

        past_inventories, past_regular_orders, past_expedited_orders = self.simulate(
            sourcing_model=sourcing_model, sourcing_periods=sourcing_periods, seed=seed
        )
        fig, ax = plt.subplots(ncols=2, figsize=(10, 4))
        ax[0].step(
            range(sourcing_periods),
            past_inventories[-sourcing_periods:],
            linewidth=linewidth,
            color="tab:blue",
        )
        ax[0].yaxis.get_major_locator().set_params(integer=True)
        ax[0].set_title("Inventory")
        ax[0].set_xlabel("Period")
        ax[0].set_ylabel("Quantity")

        ax[1].step(
            range(sourcing_periods),
            past_expedited_orders[-sourcing_periods:],
            label="Expedited Order",
            linewidth=linewidth,
            color="tab:green",
        )
        ax[1].step(
            range(sourcing_periods),
            past_regular_orders[-sourcing_periods:],
            label="Regular Order",
            linewidth=linewidth,
            color="tab:orange",
        )
        ax[1].yaxis.get_major_locator().set_params(integer=True)
        ax[1].set_title("Order")
        ax[1].set_xlabel("Period")
        ax[1].set_ylabel("Quantity")
        ax[1].legend()

    def save(self, path):
        """
        Save the controller to the specified path.
        """
        import pickle
        
        pickle.dump(self, path)

    def load(self, path):
        """
        Load the controller from the specified path.
        """
        import pickle

        pickle.load(path)


import pytest
import torch

from idinn.demand import UniformDemand
from idinn.sourcing_model import DualSourcingModel


@pytest.fixture
def dual_sourcing_model():
    return DualSourcingModel(
        regular_lead_time=2,
        expedited_lead_time=1,
        regular_order_cost=0,
        expedited_order_cost=20,
        holding_cost=0.5,
        shortage_cost=1,
        init_inventory=10,
        demand_generator=UniformDemand(low=1, high=4),
        batch_size=3,
    )


def test_dual_sourcing_model_initialization(dual_sourcing_model: DualSourcingModel):
    assert (
        dual_sourcing_model.get_regular_lead_time() == 2
    ), "Regular lead time should be 2."
    assert (
        dual_sourcing_model.get_expedited_lead_time() == 1
    ), "Expedited lead time should be 1."
    assert (
        dual_sourcing_model.get_regular_order_cost() == 0
    ), "Regular order cost should be 0."
    assert (
        dual_sourcing_model.get_expedited_order_cost() == 20
    ), "Expedited order cost should be 20."
    assert torch.all(
        dual_sourcing_model.get_init_inventory() == torch.tensor([10.0])
    ), "Initial inventory should be 10."


def test_dual_sourcing_model_order(dual_sourcing_model: DualSourcingModel):
    # Test order function for both regular and expedited orders
    regular_orders = torch.tensor([[5.0], [6.0], [7.0]])
    expedited_orders = torch.tensor([[2.0], [2.0], [2.0]])
    dual_sourcing_model.order(regular_orders, expedited_orders, seed=42)
    assert dual_sourcing_model.get_past_regular_orders().shape == (
        3,
        2,
    ), "Regular orders should be updated correctly."
    assert dual_sourcing_model.get_past_expedited_orders().shape == (
        3,
        2,
    ), "Expedited orders should be updated correctly."
    assert torch.all(
        torch.eq(dual_sourcing_model.get_last_regular_order(), regular_orders)
    ), "Last regular orders do not match expected values."
    assert torch.all(
        torch.eq(dual_sourcing_model.get_last_expedited_order(), expedited_orders)
    ), "Last expedited orders do not match expected values."

    # Test inventory updates with regular and expedited orders
    current_inventory = dual_sourcing_model.get_current_inventory()
    expected_inventory = torch.tensor([[6.0], [6.0], [8.0]])
    assert torch.all(
        current_inventory == expected_inventory
    ), f"Expected inventory {expected_inventory}, got {current_inventory}."


def test_dual_sourcing_model_multiple_periods(dual_sourcing_model: DualSourcingModel):
    # Test demand tracking across multiple periods
    for _ in range(3):
        dual_sourcing_model.order(
            torch.tensor([[1.0], [1.0], [1.0]]),
            torch.tensor([[1.0], [1.0], [1.0]]),
            seed=42,
        )
    past_demands = dual_sourcing_model.get_past_demands()
    assert past_demands.shape == (
        3,
        4,
    ), f"Expected demand shape (3, 4), got {past_demands.shape}."
    assert torch.all(past_demands[:, -1] >= 1) and torch.all(
        past_demands[:, -1] <= 4
    ), "Demands should respect the generator's range."


def test_dual_sourcing_model_reset(dual_sourcing_model: DualSourcingModel):
    # Test the reset function
    dual_sourcing_model.order(
        torch.tensor([[2.0], [3.0], [4.0]]), torch.tensor([[1.0], [1.0], [1.0]])
    )
    dual_sourcing_model.reset(batch_size=5)
    assert (
        dual_sourcing_model.batch_size == 5
    ), "Batch size should be updated to 5 after reset."
    assert dual_sourcing_model.get_past_regular_orders().shape == (
        5,
        1,
    ), "Past regular orders should be reset."
    assert dual_sourcing_model.get_past_expedited_orders().shape == (
        5,
        1,
    ), "Past expedited orders should be reset."
    assert dual_sourcing_model.get_past_inventories().shape == (
        5,
        1,
    ), "Past inventories should be reset to initial values."



import pytest

from idinn.demand import UniformDemand
from idinn.dual_controller import DynamicProgrammingController
from idinn.sourcing_model import DualSourcingModel


@pytest.fixture
def dual_sourcing_model_dp():
    return DualSourcingModel(
        regular_lead_time=2,
        expedited_lead_time=0,
        regular_order_cost=0,
        expedited_order_cost=20,
        holding_cost=5,
        shortage_cost=495,
        init_inventory=0,
        demand_generator=UniformDemand(low=0, high=4),
    )


@pytest.fixture(scope="module")
def trained_dp_controller():
    dual_sourcing_model_train = DualSourcingModel(
        regular_lead_time=2,
        expedited_lead_time=0,
        regular_order_cost=0,
        expedited_order_cost=20,
        holding_cost=5,
        shortage_cost=495,
        init_inventory=0,
        demand_generator=UniformDemand(low=0, high=4),
    )
    controller_dp = DynamicProgrammingController()
    controller_dp.fit(dual_sourcing_model_train, max_iterations=101, tolerance=1, validation_freq=100,)
    return controller_dp


def test_dp_controller_avg_cost(dual_sourcing_model_dp, trained_dp_controller):
    # Calculate average cost
    avg_cost = trained_dp_controller.get_average_cost(
        dual_sourcing_model_dp, sourcing_periods=1000, seed=42
    )
    assert abs(avg_cost - 24) < 1, f"Average cost should be near 24, but got {avg_cost}"


def test_dp_controller_plot(dual_sourcing_model_dp, trained_dp_controller):
    # Ensure plotting does not raise errors
    trained_dp_controller.plot(dual_sourcing_model_dp, sourcing_periods=100)


def test_dp_controller_order_prediction(trained_dp_controller):
    # Predict orders using the trained controller
    regular_order, expedited_order = trained_dp_controller.predict(
        current_inventory=3, past_regular_orders=[1], past_expedited_orders=[0]
    )

    # Validate predictions
    assert (
        regular_order == 3
    ), "Predicted regular order should be 3, but got {regular_order}."
    assert (
        expedited_order == 1
    ), "Predicted expedited order should be 1, but got {expedited_order}."

def test_cdi_controller_simulate(dual_sourcing_model_dp, trained_dp_controller):
    # Simulate
    past_inventories, past_regular_orders, past_expedited_orders = (
        trained_dp_controller.simulate(
            dual_sourcing_model_dp, sourcing_periods=100, seed=42
        )
    )

    # Validate simulation results
    assert (
        len(past_inventories) == 101
    ), "Simulation did not return correct number of inventory records."
    assert (
        len(past_regular_orders) == 101
    ), "Simulation did not return correct number of regular order records."
    assert (
        len(past_expedited_orders) == 101
    ), "Simulation did not return correct number of expedited order records."



import torch
from typing import Optional, Union

from ..sourcing_model import SingleSourcingModel
from .base import BaseSingleController

import logging
from datetime import datetime

# Add logger setup at class level
logger = logging.getLogger(__name__)


class BaseStockController(BaseSingleController):
    """
    Base stock controller for single-sourcing inventory optimization.
    """

    def __init__(self) -> None:
        self.sourcing_model: Optional[SingleSourcingModel] = None
        self.z_star: Optional[int] = None
        logger.info("Initialized BaseStockController")

    def fit(
        self,
        sourcing_model: SingleSourcingModel,
        num_samples: int = 100000,
        seed: Optional[int] = None,
    ) -> None:
        """
        Calculate the optimal target inventory level z* and store it in self.z_star.

        Parameters
        ----------
        sourcing_model : SourcingModel
            The sourcing model for training.
        num_samples : int
            The number of samples used for calculating empirical percentile.
        seed : int, optional
            Random seed for reproducibility.
        """
        # Store sourcing model in self.sourcing_model
        self.sourcing_model = sourcing_model

        if seed is not None:
            torch.manual_seed(seed)

        start_time = datetime.now()
        logger.info(f"Starting base stock policy calculation at {start_time}")
        logger.info(f"Sourcing model parameters: batch_size={self.sourcing_model.batch_size}, "
                    f"lead_time={self.sourcing_model.lead_time}, init_inventory={self.sourcing_model.init_inventory.int().item()}, "
                    f"demand_generator={self.sourcing_model.demand_generator.__class__.__name__}")
        logger.info(f"Training parameters: num_samples={num_samples}")

        # Generate samples for l + 1 periods
        samples = sourcing_model.demand_generator.sample(
            batch_size=num_samples, batch_width=self.sourcing_model.get_lead_time() + 1
        )

        # Calculate the total demand for each sample
        total_demand_samples = samples.sum(dim=1)

        # Get shortage cost and holding cost from sourcing model
        b = sourcing_model.get_shortage_cost()
        h = sourcing_model.get_holding_cost()

        # Calculate z* using the empirical percentile (inverse CDF)
        service_level = b / (b + h)
        self.z_star = (
            torch.quantile(total_demand_samples.float(), service_level).int().item()
        )

        end_time = datetime.now()
        duration = end_time - start_time
        logger.info(f"Policy calculation completed at {end_time}")
        logger.info(f"Total calculation duration: {duration}")
        logger.info(f"Optimal base stock level (z*): {self.z_star}")
        logger.info(f"Final best cost: {self.get_average_cost(self.sourcing_model, sourcing_periods=1000, seed=42):.4f}")

    def predict(
        self,
        current_inventory: Union[int, torch.Tensor],
        past_orders: Optional[Union[list, torch.Tensor]] = None,
        output_tensor: bool = False,
    ) -> Union[torch.Tensor, int]:
        """
        Calculate the replenishment order quantity.

        Parameters
        ----------
        current_inventory : int
            Current inventory level.
        past_orders : list, or torch.Tensor, optional
            Array of past orders. If `past_orders` is None, or the length of `past_orders` is lower than `lead_time`, it will be padded with zeros. If the length of `past_orders` is higher than `lead_time`, only the last `lead_time` orders will be used during inference.
        output_tensor : bool, default is False
            If True, the replenishment order quantity will be returned as a torch.Tensor. Otherwise, it will be returned as an integer.

        Returns
        -------
        float
            The replenishment order quantity.
        """
        if self.sourcing_model is None:
            raise AttributeError("The controller is not trained.")

        lead_time = self.sourcing_model.get_lead_time()

        current_inventory = self._current_inventory_check(current_inventory)
        past_orders = self._past_orders_check(past_orders, lead_time)

        if lead_time == 0:
            inventory_position = current_inventory
        elif lead_time > 0:
            inventory_position = current_inventory + past_orders[:, -lead_time:].sum(
                dim=1, keepdim=True
            )
        else:
            raise ValueError("`lead_time` cannot be less than 0")

        result = torch.relu(self.z_star - inventory_position)

        if output_tensor:
            return result
        else:
            return result.int().item()

    def reset(self) -> None:
        self.z_star = None
        self.sourcing_model = None



import pytest
import torch

from idinn.demand import CustomDemand
from idinn.sourcing_model import DualSourcingModel


@pytest.fixture
def dual_sourcing_model_custom_demand():
    """
    Dual sourcing model with fixed demand at 1.
    """
    return DualSourcingModel(
        regular_lead_time=1,
        expedited_lead_time=0,
        regular_order_cost=0,
        expedited_order_cost=20,
        holding_cost=0.5,
        shortage_cost=1,
        init_inventory=10,
        batch_size=3,
        demand_generator=CustomDemand({1: 1.0}),
    )


def test_custom_demand_initialization(
    dual_sourcing_model_custom_demand: DualSourcingModel,
):
    # Test that the model initializes correctly
    assert (
        dual_sourcing_model_custom_demand.get_regular_lead_time() == 1
    ), "Regular lead time should be 1."
    assert (
        dual_sourcing_model_custom_demand.get_expedited_lead_time() == 0
    ), "Expedited lead time should be 0."
    assert torch.all(
        dual_sourcing_model_custom_demand.get_init_inventory() == torch.tensor([10.0])
    ), "Initial inventory should be 10."


def test_custom_demand_order(
    dual_sourcing_model_custom_demand: DualSourcingModel,
):
    # Test that the demand generator consistently outputs 1.0
    dual_sourcing_model_custom_demand.order(
        torch.tensor([[3.0], [3.0], [3.0]]), torch.tensor([[1.0], [1.0], [1.0]])
    )
    past_demands = dual_sourcing_model_custom_demand.get_past_demands()
    assert torch.all(
        past_demands[:, -1] == 1.0
    ), "Custom demand should consistently output a fixed value of 1.0."

    # Test that inventories update correctly with regular and expedited orders
    dual_sourcing_model_custom_demand.order(
        torch.tensor([[1.0], [1.0], [1.0]]), torch.tensor([[1.0], [1.0], [1.0]])
    )
    current_inventory = dual_sourcing_model_custom_demand.get_current_inventory()
    expected_inventory = torch.tensor(
        [[13.0], [13.0], [13.0]]
    )  # 10 initial + 3 regular + 2 expedited - 2 demand
    assert torch.all(
        current_inventory == expected_inventory
    ), f"Expected inventory {expected_inventory}, got {current_inventory}."


def test_custom_demand_past_orders(
    dual_sourcing_model_custom_demand: DualSourcingModel,
):
    # Test that both regular and expedited past orders are tracked correctly
    regular_orders = torch.tensor([[5.0], [6.0], [7.0]])
    expedited_orders = torch.tensor([[1.0], [1.0], [1.0]])
    dual_sourcing_model_custom_demand.order(regular_orders, expedited_orders)
    past_regular_orders = dual_sourcing_model_custom_demand.get_past_regular_orders()
    past_expedited_orders = (
        dual_sourcing_model_custom_demand.get_past_expedited_orders()
    )
    assert torch.all(
        past_regular_orders[:, -1] == regular_orders.squeeze()
    ), "Last regular orders do not match expected values."
    assert torch.all(
        past_expedited_orders[:, -1] == expedited_orders.squeeze()
    ), "Last expedited orders do not match expected values."


def test_custom_demand_multiple_periods(
    dual_sourcing_model_custom_demand: DualSourcingModel,
):
    # Test model behavior over multiple periods
    for _ in range(5):
        dual_sourcing_model_custom_demand.order(
            torch.tensor([[2.0], [2.0], [2.0]]), torch.tensor([[1.0], [1.0], [1.0]])
        )
    past_inventories = dual_sourcing_model_custom_demand.get_past_inventories()
    past_demands = dual_sourcing_model_custom_demand.get_past_demands()
    assert past_inventories.shape == (
        3,
        6,
    ), f"Expected inventory shape (3, 6), got {past_inventories.shape}."
    assert past_demands.shape == (
        3,
        6,
    ), f"Expected demand shape (3, 6), got {past_demands.shape}."
    assert torch.all(
        past_demands[:, -1] == 1.0
    ), "Custom demand should consistently output a fixed value of 1.0."


def test_custom_demand_reset(dual_sourcing_model_custom_demand: DualSourcingModel):
    # Test resetting the model
    dual_sourcing_model_custom_demand.order(
        torch.tensor([[2.0], [2.0], [2.0]]), torch.tensor([[1.0], [1.0], [1.0]])
    )
    dual_sourcing_model_custom_demand.reset()
    assert torch.all(
        dual_sourcing_model_custom_demand.get_past_regular_orders() == torch.zeros(3, 1)
    ), "Past regular orders should be reset to zero."
    assert torch.all(
        dual_sourcing_model_custom_demand.get_past_expedited_orders()
        == torch.zeros(3, 1)
    ), "Past expedited orders should be reset to zero."
    assert torch.all(
        dual_sourcing_model_custom_demand.get_past_demands() == torch.zeros(3, 1)
    ), "Past demands should be reset to zero."
    assert torch.all(
        dual_sourcing_model_custom_demand.get_past_inventories()
        == torch.tensor([10.0]).repeat(3, 1)
    ), "Past inventories should reset to initial inventory."



from typing import List, Optional, Tuple, Union

import torch

from ..sourcing_model import DualSourcingModel
from .base import BaseDualController

import logging
from datetime import datetime

# Get root logger
logger = logging.getLogger()

class CappedDualIndexController(BaseDualController):
    """
    Controller class for capped dual index inventory optimization.

    Parameters
    ----------
    s_e : int
        Capped dual index parameter 1
    s_r : int
        Capped dual index parameter 2
    q_r : int
        Capped dual index parameter 3

    Notes
    -----
    The function follows the implementation of Sun, J., & Van Mieghem, J. A. (2019)([1]_).

    References
    ----------
    .. [1] Robust dual sourcing inventory management: Optimality of capped dual index policies and smoothing.
           Manufacturing & Service Operations Management, 21(4), 912-931.
    """

    def __init__(self, s_e: int = 0, s_r: int = 0, q_r: int = 0) -> None:
        self.sourcing_model = None
        self.s_e = s_e
        self.s_r = s_r
        self.q_r = q_r
        logger.info("Initialized CappedDualIndexController")

    def capped_dual_index_sum(
        self,
        current_inventory: int,
        past_regular_orders: torch.Tensor,
        past_expedited_orders: torch.Tensor,
        limit: bool = False,
    ) -> int:
        """
        Calculate the capped dual index sum.

        Parameters
        ----------
        current_inventory : int
            Current inventory level.
        past_regular_orders : torch.Tensor
            Array of past regular orders.
        past_expedited_orders : torch.Tensor
            Array of past expedited orders.
        regular_lead_time : int
            Regular lead time.
        expedited_lead_time : int
            Expedited lead time.
        limit : bool
            If true, set parameter k for capped dual index sum calculation, where 0 <= k <= l_r -1,
            to regular_lead_time - expedited_lead_time - 1. Else, set k to 0.

        Returns
        -------
        int
            The capped dual index sum.
        """
        if self.sourcing_model is None:
            raise AttributeError("The controller is not trained.")

        regular_lead_time = self.sourcing_model.get_regular_lead_time()
        expedited_lead_time = self.sourcing_model.get_expedited_lead_time()

        current_inventory = self._check_current_inventory(current_inventory)
        past_regular_orders = self._check_past_orders(
            past_regular_orders, regular_lead_time
        )
        past_expedited_orders = self._check_past_orders(
            past_expedited_orders, expedited_lead_time
        )

        if limit:
            k = regular_lead_time - expedited_lead_time - 1
        else:
            k = 0

        inventory_position = (
            current_inventory
            + past_regular_orders[
                :, -regular_lead_time : -regular_lead_time + k + 1
            ].sum()
        )

        if expedited_lead_time >= 1:
            inventory_position += past_expedited_orders[
                -expedited_lead_time : +min(k - expedited_lead_time, -1) + 1
            ].sum()

        return inventory_position

    def fit(
        self,
        sourcing_model: DualSourcingModel,
        sourcing_periods: int,
        s_e_range: torch.Tensor = torch.arange(2, 11),
        s_r_range: torch.Tensor = torch.arange(2, 11),
        q_r_range: torch.Tensor = torch.arange(2, 11),
        seed: Optional[int] = None,
    ) -> None:
        """
        Train the capped dual index controller.

        Parameters
        ----------
        sourcing_model : SourcingModel
            The sourcing model.
        sourcing_periods : int
            Number of sourcing periods.
        s_e_range : torch.Tensor, optional
            Range of values for s_e.
        s_r_range : torch.Tensor, optional
            Range of values for s_r.
        q_r_range : torch.Tensor, optional
            Range of values for q_r.
        seed : int, optional
            Random seed for reproducibility.
        """
        self.sourcing_model = sourcing_model

        if seed is not None:
            torch.manual_seed(seed)
        
        start_time = datetime.now()
        logger.info(f"Starting capped dual index grid search at {start_time}")
        logger.info(f"Sourcing model parameters: batch_size={self.sourcing_model.batch_size}, "
                    f"lead_time={self.sourcing_model.lead_time}, init_inventory={self.sourcing_model.init_inventory.int().item()}, "
                    f"demand_generator={self.sourcing_model.demand_generator.__class__.__name__}")
        logger.info(f"Training parameters: s_e_range={s_e_range.tolist()}, s_r_range={s_r_range.tolist()}, q_r_range={q_r_range.tolist()}")

        min_cost = torch.inf
        for s_e in s_e_range:
            for s_r in s_r_range:
                for q_r in q_r_range:
                    sourcing_model.reset()
                    self.s_e = s_e
                    self.s_r = s_r
                    self.q_r = q_r
                    total_cost = self.get_total_cost(sourcing_model, sourcing_periods)
                    logger.info(f"s_e={s_e}, s_r={s_r}, q_r={q_r}"
                                f" - Training cost: {total_cost/sourcing_periods:.4f}")

                    if total_cost < min_cost:
                        min_cost = total_cost
                        s_e_optimal = s_e
                        s_r_optimal = s_r
                        q_r_optimal = q_r

        self.s_e = s_e_optimal
        self.s_r = s_r_optimal
        self.q_r = q_r_optimal

        end_time = datetime.now()
        duration = end_time - start_time
        logger.info(f"Grid search completed at {end_time}")
        logger.info(f"Total training duration: {duration}")
        logger.info(f"Final best cost: {min_cost/sourcing_periods:.4f}")
        logger.info(f"Final best parameters: s_e={s_e_optimal}, s_r={s_r_optimal}, q_r={q_r_optimal}")

    def predict(
        self,
        current_inventory: int,
        past_regular_orders: Optional[torch.Tensor] = None,
        past_expedited_orders: Optional[torch.Tensor] = None,
        output_tensor: bool = False
    ) -> Tuple[Union[torch.Tensor, int], Union[torch.Tensor, int]]:
        """
        Perform forward calculation for capped dual index optimization.

        Parameters
        ----------
        current_inventory : int, or torch.Tensor
            Current inventory.
        past_regular_orders : list, or torch.Tensor, optional
            Past regular orders. If the length of `past_regular_orders` is lower than `regular_lead_time`, it will be padded with zeros. If the length of `past_regular_orders` is higher than `regular_lead_time`, only the last `regular_lead_time` orders will be used during inference.
        past_expedited_orders : list, or torch.Tensor, optional
            Past expedited orders. If the length of `past_expedited_orders` is lower than `expedited_lead_time`, it will be padded with zeros. If the length of `past_expedited_orders` is higher than `expedited_lead_time`, only the last `expedited_lead_time` orders will be used during inference.
        output_tensor : bool, default is False
            If True, the replenishment order quantity will be returned as a torch.Tensor. Otherwise, it will be returned as an integer.

        Returns
        -------
        tuple
            A tuple containing the regular order quantity and expedited order quantity.
        """
        inventory_position = self.capped_dual_index_sum(
            current_inventory,
            past_regular_orders,
            past_expedited_orders,
            limit=False,
        )
        inventory_position_lm1 = self.capped_dual_index_sum(
            current_inventory,
            past_regular_orders,
            past_expedited_orders,
            limit=True,
        )
        regular_q = int(min(max(0, self.s_r - inventory_position_lm1), self.q_r))
        expedited_q = int(max(0, self.s_e - inventory_position))

        if output_tensor:
            return torch.tensor([[regular_q]]), torch.tensor([[expedited_q]])
        else:
            return regular_q, expedited_q

    def reset(self) -> None:
        """
        Reset the controller to the initial state.
        """
        self.s_e = 0
        self.s_r = 0
        self.q_r = 0
        self.sourcing_model = None



import pytest

from idinn.demand import UniformDemand
from idinn.single_controller import BaseStockController
from idinn.sourcing_model import SingleSourcingModel


@pytest.fixture
def base_stock_controller():
    return BaseStockController()


@pytest.mark.parametrize(
    "lead_time, expected_z_star, expected_avg_cost",
    [
        (0, 4, 10),
        (2, 11, 29),
    ],
)
def test_base_stock_controller_fit_and_cost(
    base_stock_controller, lead_time, expected_z_star, expected_avg_cost
):
    single_sourcing_model = SingleSourcingModel(
        lead_time=lead_time,
        holding_cost=5,
        shortage_cost=495,
        batch_size=32,
        init_inventory=10,
        demand_generator=UniformDemand(low=0, high=4),
    )
    base_stock_controller.fit(single_sourcing_model, seed=42)
    assert (
        base_stock_controller.z_star == expected_z_star
    ), f"Expected z_star to be {expected_z_star}, but got {base_stock_controller.z_star}"

    avg_cost = base_stock_controller.get_average_cost(
        single_sourcing_model, sourcing_periods=1000, seed=42
    )
    assert (
        abs(avg_cost - expected_avg_cost) < 1
    ), f"Average cost should be near {expected_avg_cost}, but got {avg_cost}"


def test_base_stock_controller_simulation(base_stock_controller):
    single_sourcing_model = SingleSourcingModel(
        lead_time=0,
        holding_cost=5,
        shortage_cost=495,
        batch_size=32,
        init_inventory=10,
        demand_generator=UniformDemand(low=0, high=4),
    )
    base_stock_controller.fit(single_sourcing_model, seed=42)

    past_inventories, past_orders = base_stock_controller.simulate(
        single_sourcing_model, sourcing_periods=100
    )
    assert (
        len(past_inventories) == 101
    ), "Simulation did not return correct inventory records"
    assert len(past_orders) == 101, "Simulation did not return correct order records"


@pytest.mark.parametrize(
    "lead_time, current_inventory, expected_predict",
    [
        (0, 10, 0),
        (2, 10, 1),
    ],
)
def test_base_stock_controller_predict(
    lead_time, current_inventory, expected_predict, base_stock_controller
):
    single_sourcing_model = SingleSourcingModel(
        lead_time=lead_time,
        holding_cost=5,
        shortage_cost=495,
        batch_size=32,
        init_inventory=10,
        demand_generator=UniformDemand(low=0, high=4),
    )
    base_stock_controller.fit(single_sourcing_model, seed=42)
    predict = base_stock_controller.predict(current_inventory=current_inventory)
    assert (
        predict == expected_predict
    ), f"Expected predict to be {expected_predict}, but got {predict}"



import inspect
import logging
import os
import queue
import shutil
import threading
import time
from logging.handlers import QueueHandler
from typing import Any, Dict, Type

import numpy as np
import pandas as pd
import plotly.express as px
import streamlit as st
import torch
from streamlit.runtime.scriptrunner_utils.script_run_context import (
    add_script_run_ctx,
    get_script_run_ctx,
)
from torch.utils.tensorboard import SummaryWriter
from torchview import draw_graph
from utils import tflog2pandas

from idinn.demand import CustomDemand, UniformDemand
from idinn.dual_controller import DualSourcingNeuralController
from idinn.sourcing_model import DualSourcingModel

if torch.cuda.is_available():
    torch.set_default_device("cuda")
    torch.classes.__path__ = [os.path.join(torch.__path__[0], torch.classes.__file__)]

st.set_page_config(layout="wide")

if "log_queue" not in st.session_state:
    st.session_state["log_queue"] = queue.Queue()

queue_handler = QueueHandler(st.session_state["log_queue"])
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
queue_handler.setFormatter(formatter)

root_logger = logging.getLogger()
root_logger.setLevel(logging.INFO)
root_logger.addHandler(queue_handler)

# Color mapping for log levels
LOG_COLORS = {
    "DEBUG": "blue",
    "INFO": "green",
    "WARNING": "orange",
    "ERROR": "red",
    "CRITICAL": "purple",
}


def process_log_queue() -> None:
    """Check queue for new logs and update session state."""
    while True:
        try:
            log = st.session_state["log_queue"].get_nowait()
            st.session_state.logs.append(log)
        except queue.Empty:
            break


def write_logs() -> None:
    """Render logs with color coding in markdown."""
    process_log_queue()
    messages = list(map(lambda log: log.msg, st.session_state.logs[-100:]))
    prev_log = None
    for log in messages:  # Show last 100 logs
        try:
            parts = log.split(" - ")
            if len(parts) >= 3:
                timestamp, level, message = parts[0], parts[1], " - ".join(parts[2:])
                color = LOG_COLORS.get(level, "gray")
                if log != prev_log:
                    st.markdown(f"`{timestamp}` :{color}[**{level}**] {message}")
                prev_log = log
        except Exception as e:
            st.write(f"Error parsing log: {str(e)}")
            time.sleep(0.1)


def update_log_display(log_placeholder: st.delta_generator.DeltaGenerator) -> None:
    """Update the log display.

    Parameters
    ----------
    log_placeholder : st.delta_generator.DeltaGenerator
        The Streamlit placeholder for logs.
    """
    with log_placeholder:
        while True:
            write_logs()
            time.sleep(1)


def initialize_session_state() -> None:
    """Initialize the session state."""
    for key in [
        "training",
        "trainingnow",
        "demand_generator",
        "dual_sourcing_model",
        "demand_controller",
    ]:
        if key not in st.session_state:
            st.session_state[key] = (
                0 if key == "training" else False if key == "trainingnow" else None
            )
    if "logs" not in st.session_state:
        st.session_state["logs"] = []


def fit_model(
    controller: DualSourcingNeuralController,
    sourcing_model: DualSourcingModel,
    sourcing_periods: int,
    validation_sourcing_periods: int,
    epochs: int,
    seed: int,
) -> DualSourcingNeuralController:
    """Fit the model.

    Parameters
    ----------
    controller : DualSourcingNeuralController
        The neural network controller.
    sourcing_model : DualSourcingModel
        The dual sourcing model.
    sourcing_periods : int
        The number of training sourcing periods.
    validation_sourcing_periods : int
        The number of validation sourcing periods.
    epochs : int
        The number of training epochs.
    seed : int
        The random seed.

    Returns
    -------
    DualSourcingNeuralController
        The fitted controller.
    """
    if torch.cuda.is_available():
        torch.set_default_device("cuda")
        torch.classes.__path__ = [
            os.path.join(torch.__path__[0], torch.classes.__file__)
        ]

    controller.fit(
        sourcing_model=sourcing_model,
        sourcing_periods=sourcing_periods,
        validation_sourcing_periods=validation_sourcing_periods,
        epochs=epochs,
        tensorboard_writer=SummaryWriter("runs/dual_sourcing_model"),
        seed=seed,
    )
    return controller


def click() -> None:
    """Callback function for training button click."""
    st.session_state["training"] += 1
    if os.path.exists("runs/dual_sourcing_model"):
        shutil.rmtree("runs/dual_sourcing_model")
    st.session_state["trainingnow"] = True

    log_placeholder = st.empty()

    # Start the logging thread
    log_thread = threading.Thread(target=update_log_display, args=(log_placeholder,))
    add_script_run_ctx(log_thread, get_script_run_ctx())
    log_thread.start()

    fit_model(
        st.session_state["demand_controller"],
        st.session_state["dual_sourcing_model"],
        sourcing_periods,
        validation_sourcing_periods,
        epochs,
        seed,
    )

    st.session_state["trainingnow"] = False
    st.success("Training complete! Check 'Result' tab.")

    # # Ensure the logging thread finishes
    # log_thread.join()


def training_fragment() -> None:
    """Render the training button."""
    cf1, cf2 = st.columns([1, 4])
    with cf1:
        pressed = st.button("Fit Controller")
    with cf2:
        if pressed:
            click()


# Session State Initialization
initialize_session_state()

# Page Header
st.header("Inventory Dynamics–Informed Neural Networks")
st.markdown(
    "Welcome to Inventory Dynamics–Informed Neural Networks! This application generates ordering policies from expedited and regular suppliers. Use the sidebar on the left to select a demand model, choose your preferred solver, and view the results after fitting."
)

# Tabs
tab1, tab2, tab3, tab4 = st.tabs(
    ["Demand Generation", "Dual Sourcing Model", "NN Controller", "Result"]
)

# Demand Generation Tab
with tab1:
    c1, c2 = st.columns([1, 3])

    with c1:
        st.subheader("Demand Generation")

        demand_type = st.radio(
            label="Please choose demand type", options=["Uniform", "File"]
        )
        if demand_type == "Uniform":
            with st.form(key="uniform_demand"):
                low = st.number_input(
                    "Minimum demand",
                    value=1,
                    step=1,
                    format="%i",
                    max_value=(1 << 53) - 1,
                )
                high = st.number_input(
                    "Maximum demand", value=4, step=1, format="%i", min_value=0
                )
                submitted = st.form_submit_button("Generate")

                if submitted and high >= low:
                    st.success(
                        f"Successfully generated uniform demand within range: [{np.floor(low)}, {np.floor(high)}]!"
                    )
                    st.session_state["training"] = 0
                    st.session_state["demand_generator"] = UniformDemand(low, high)
                elif submitted:
                    st.error(
                        "Please resubmit and make sure that maximum demand is greater or equal to minimum demand."
                    )
                    st.session_state["training"] = 0

        elif demand_type == "File":
            with st.form(key="uniform_demand"):
                uploaded_file = st.file_uploader(
                    label="Please upload a single column file with demand values. Each row represents a timestep and each element represents a demand value."
                )
                submitted2 = st.form_submit_button("Generate")
                if submitted2:
                    try:
                        df = pd.read_csv(uploaded_file)
                        st.success(
                            f"Successfully uploaded file and contains {df.shape[0]} demand points!"
                        )
                        st.session_state["demand_generator"] = CustomDemand(
                            torch.tensor(df.iloc[:, 0].values)
                        )
                        st.session_state["training"] = 0
                    except Exception:
                        st.error("Could not parse file! Please try again!")

    with c2:
        if st.session_state["demand_generator"] is not None:
            all_demands = [
                st.session_state["demand_generator"].sample(1).item()
                for _ in range(100)
            ]
            c2c1, c2c2 = st.columns(2)
            with c2c1:
                fig = px.line(y=all_demands, line_shape="hv").update_layout(
                    xaxis_title="Periods",
                    yaxis_title="Demand",
                    title="Typical Demand Trajectory",
                )
                st.plotly_chart(fig, use_container_width=True)
            with c2c2:
                fig = px.histogram(x=all_demands).update_layout(
                    xaxis_title="Demand",
                    yaxis_title="Frequency",
                    title="Demand Distribution",
                )
                st.plotly_chart(fig, use_container_width=True)

# Dual Sourcing Model Tab
with tab2:
    st.subheader("Dual Sourcing Model")
    with st.form(key="Dual Sourcing Model"):
        c1t2, c2t2 = st.columns(2)
        with c1t2:
            regular_lead_time = np.int32(
                st.number_input(
                    "Regular lead time:", value=2, min_value=0, format="%i", step=1
                )
            )
            expedited_lead_time = np.int32(
                st.number_input(
                    "Expedited lead time:", value=0, min_value=0, format="%i", step=1
                )
            )
            batch_size = np.int32(
                st.number_input(
                    "Minibatch size for demand trajectories:",
                    value=16,
                    min_value=0,
                    format="%i",
                    step=1,
                )
            )
            init_inventory = np.int32(
                st.number_input(
                    "Initial inventory:", value=6, min_value=0, format="%i", step=1
                )
            )

        with c2t2:
            regular_order_cost = np.int32(
                st.number_input(
                    "Regular order cost:", value=0, min_value=0, format="%i", step=1
                )
            )
            expedited_order_cost = np.int32(
                st.number_input(
                    "Expedited order cost:", value=20, min_value=0, format="%i", step=1
                )
            )
            holding_cost = np.int32(
                st.number_input(
                    "Holding cost:", value=5, min_value=0, format="%i", step=1
                )
            )
            shortage_cost = np.int32(
                st.number_input(
                    "Shortage cost:", value=495, min_value=0, format="%i", step=1
                )
            )

        model_params = {
            "regular_lead_time": regular_lead_time,
            "expedited_lead_time": expedited_lead_time,
            "regular_order_cost": regular_order_cost,
            "expedited_order_cost": expedited_order_cost,
            "holding_cost": holding_cost,
            "shortage_cost": shortage_cost,
            "batch_size": batch_size,
            "init_inventory": init_inventory,
            "demand_generator": st.session_state["demand_generator"],
        }

        cc1, cc2 = st.columns([1, 4])
        with cc1:
            pressed2 = st.form_submit_button("Create Sourcing Model")
        with cc2:
            if pressed2:
                st.session_state["training"] = 0
                st.session_state["dual_sourcing_model"] = DualSourcingModel(
                    **model_params
                )
                st.success("Successfully created sourcing model!")

# Controller Definition Tab
with tab3:
    st.subheader("Controller Definition")
    c1, c2 = st.columns(2)
    with c1:
        activation_map: Dict[str, Type[torch.nn.Module]] = {
            "CELU": torch.nn.CELU,
            "ReLU": torch.nn.ReLU,
            "LeakyReLU": torch.nn.LeakyReLU,
            "ELU": torch.nn.ELU,
            "Tanh": torch.nn.Tanh,
            "Sigmoid": torch.nn.Sigmoid,
            "SiLU": torch.nn.SiLU,
            "GELU": torch.nn.GELU,
            "TahnShrink": torch.nn.Tanhshrink,
        }
        activation_id: str = st.selectbox(
            "Activation function of hidden layers:", options=list(activation_map.keys())
        )
        activation_signature: inspect.Signature = inspect.signature(
            activation_map[activation_id]
        )
        kwargs: Dict[str, Any] = {
            v.name: st.number_input(
                f"Value for activation function's parameter: {v.name}",
                value=(v.default if v.default is not None else 0.0),
            )
            for k, v in activation_signature.parameters.items()
            if v.annotation in [float, int]
        }

        layer_sizes: str = st.text_area(
            label="Layer sizes:",
            value="128, 64, 32, 16, 8, 4",
            help="A comma separated list of integers, indicating neurons per layer, starting from the first hidden layer (leftmost)",
        )
        sourcing_periods: int = st.number_input(
            "Number of training sourcing periods:",
            value=50,
            min_value=1,
            format="%i",
            step=1,
        )
        validation_sourcing_periods: int = st.number_input(
            "Number of validation sourcing periods:",
            value=1000,
            min_value=1,
            format="%i",
            step=1,
        )
        epochs: int = st.number_input(
            "Number of training epochs:", value=2000, min_value=1, format="%i", step=1
        )
        seed: int = st.number_input(
            "Seed:", value=1234, min_value=1, format="%i", step=1
        )
        try:
            layer_sizes = list(map(int, layer_sizes.split(",")))
        except Exception:
            st.error("Provided input cannot be parsed to layers.")

        st.session_state["demand_controller"] = DualSourcingNeuralController(
            hidden_layers=layer_sizes,
            activation=activation_map[activation_id](**kwargs),
        )

    with c2:
        x = torch.linspace(-10, 10, 100)
        fig = px.line(
            x=x.cpu().numpy(),
            y=activation_map[activation_id](**kwargs)(x).cpu().numpy(),
            title=f"Activation shape: {activation_id}",
        )
        st.plotly_chart(fig, use_container_width=True)

        if (
            st.session_state["demand_controller"]
            and st.session_state["dual_sourcing_model"]
        ):
            st.info(
                "Graph plot for minibatch size of 4 (for illustration purposes). Click top right corner to enlarge!"
            )
            st.session_state["demand_controller"].init_layers(
                regular_lead_time=regular_lead_time,
                expedited_lead_time=expedited_lead_time,
            )
            pre_input_tensors = (
                torch.rand([4, 1]),
                torch.rand([4, max(regular_lead_time, 1)]),
                torch.rand([4, max(regular_lead_time, 1)]),
            )

            input_tensor = st.session_state["demand_controller"].prepare_inputs(
                *pre_input_tensors,
                sourcing_model=st.session_state["dual_sourcing_model"],
            )

            input_sizes = [input_tensor.shape]
            model_graph = draw_graph(
                st.session_state["demand_controller"].model, input_size=input_sizes
            )
            model_graph.visual_graph.attr("graph", rankdir="LR")
            st.graphviz_chart(model_graph.visual_graph)
        elif not st.session_state["dual_sourcing_model"]:
            st.warning(
                "Please define a dual sourcing model to generate NN architecture graph!"
            )

    log_placeholder = st.empty()
    training_fragment()

# Results Tab
with tab4:
    if all(
        st.session_state[key]
        for key in ["demand_controller", "dual_sourcing_model", "demand_generator"]
    ):
        if (
            os.path.exists("runs/dual_sourcing_model")
            and st.session_state["training"] > 0
        ):
            try:
                t4c1, t4c2 = st.columns(2)
                with t4c1:
                    tsb_df = tflog2pandas("runs/dual_sourcing_model")
                    # Identify and separate outliers
                    threshold = (
                        tsb_df["value"].quantile(0.99) * 10
                    )  # Define a threshold for outliers
                    outliers = tsb_df[tsb_df["value"] > threshold]
                    inliers = tsb_df[tsb_df["value"] <= threshold]

                    fig = px.line(
                        inliers,
                        x="step",
                        y="value",
                        color="metric",
                        line_shape="hv",
                        title="Learning Curves",
                    ).update_layout(
                        yaxis_title="Avg. Cost per Period", xaxis_title="Epoch"
                    )

                    # Add outliers to the plot
                    if not outliers.empty:
                        fig.add_scatter(
                            x=outliers["step"],
                            y=outliers["value"],
                            mode="markers",
                            name="Outliers",
                            color="metric",
                        )

                    st.plotly_chart(fig)
                with t4c2:
                    past_inventories, past_regular_orders, past_expedited_orders = (
                        st.session_state["demand_controller"].simulate(
                            sourcing_model=st.session_state["dual_sourcing_model"],
                            sourcing_periods=sourcing_periods,
                        )
                    )
                    df_past = pd.DataFrame(
                        {
                            "Inventory": past_inventories,
                            "Regular Orders": past_regular_orders,
                            "Expedited Orders": past_expedited_orders,
                        }
                    )
                    fig = px.line(df_past, line_shape="hv").update_layout(
                        xaxis_title="Periods",
                        yaxis_title="Quantity",
                        title="Sample Optimization Trajectory",
                    )
                    st.plotly_chart(fig)
            except Exception as e:
                st.warning(
                    "No available model, please make sure you submit the previous steps!"
                )
                st.error(e)
    else:
        if not st.session_state["demand_generator"]:
            st.warning(
                "No demand generator is chosen for the current configuration, please define one."
            )
        if not st.session_state["dual_sourcing_model"]:
            st.warning(
                "No sourcing model is defined for the current configuration, please define one."
            )
        if not st.session_state["demand_controller"]:
            st.warning(
                "No model chosen or trained for the current configuration, please go to the previous step and fit a model."
            )



import pytest
import torch

from idinn.demand import CustomDemand
from idinn.sourcing_model import SingleSourcingModel


@pytest.fixture
def single_sourcing_model_custom_demand():
    """
    Single sourcing model with fixed demand at 1.
    """
    return SingleSourcingModel(
        lead_time=1,
        holding_cost=0.5,
        shortage_cost=1,
        init_inventory=10,
        batch_size=3,
        demand_generator=CustomDemand({1: 1.0}),
    )


def test_custom_demand_initialization(
    single_sourcing_model_custom_demand: SingleSourcingModel,
):
    assert (
        single_sourcing_model_custom_demand.get_lead_time() == 1
    ), "Lead time should be 1."
    assert torch.all(
        single_sourcing_model_custom_demand.get_init_inventory() == torch.tensor([10.0])
    ), "Initial inventory should be 10."


def test_custom_demand_order(
    single_sourcing_model_custom_demand: SingleSourcingModel,
):
    # Simulate orders and ensure inventory updates correctly with demand fixed at 1
    single_sourcing_model_custom_demand.order(torch.tensor([[6.0], [6.0], [6.0]]))
    single_sourcing_model_custom_demand.order(torch.tensor([[0.0], [0.0], [0.0]]))
    current_inventory = single_sourcing_model_custom_demand.get_current_inventory()
    expected_inventory = torch.tensor(
        [[14.0], [14.0], [14.0]]
    )  # 10 initial + 6 orders - 2 demand
    assert torch.all(
        current_inventory == expected_inventory
    ), f"Expected inventory {expected_inventory}, but got {current_inventory}."
    past_demands = single_sourcing_model_custom_demand.get_past_demands()
    assert torch.all(
        past_demands[:, -1] == 1
    ), "Custom demand should always generate a demand of 1."


def test_custom_demand_past_orders(
    single_sourcing_model_custom_demand: SingleSourcingModel,
):
    single_sourcing_model_custom_demand.order(torch.tensor([[2.0], [3.0], [4.0]]))
    past_orders = single_sourcing_model_custom_demand.get_past_orders()
    assert torch.all(
        torch.eq(past_orders[:, -1], torch.tensor([2.0, 3.0, 4.0]))
    ), f"Past orders should match the last orders made. Got {past_orders[:, -1]}."


def test_custom_demand_multiple_periods(
    single_sourcing_model_custom_demand: SingleSourcingModel,
):
    for _ in range(5):
        single_sourcing_model_custom_demand.order(
            torch.tensor([[1.0], [1.0], [1.0]]), seed=42
        )
    past_demands = single_sourcing_model_custom_demand.get_past_demands()
    past_inventories = single_sourcing_model_custom_demand.get_past_inventories()
    assert past_demands.shape == (
        3,
        6,
    ), f"Expected demand shape (3, 6), got {past_demands.shape}."
    assert past_inventories.shape == (
        3,
        6,
    ), f"Expected inventory shape (3, 6), got {past_inventories.shape}."
    assert torch.all(
        past_demands[:, -1] == 1
    ), "Custom demand should generate a demand of 1 for all periods."


def test_custom_demand_reset(single_sourcing_model_custom_demand: SingleSourcingModel):
    # Test resetting the model
    single_sourcing_model_custom_demand.order(
        torch.tensor([[2.0], [2.0], [2.0]]), seed=42
    )
    single_sourcing_model_custom_demand.reset()
    assert torch.all(
        single_sourcing_model_custom_demand.get_past_orders() == torch.zeros(3, 1)
    ), "Past orders should be reset to zero."
    assert torch.all(
        single_sourcing_model_custom_demand.get_past_demands() == torch.zeros(3, 1)
    ), "Past demands should be reset to zero."
    assert torch.all(
        single_sourcing_model_custom_demand.get_past_inventories()
        == torch.tensor([10.0]).repeat(3, 1)
    ), "Past inventories should reset to initial inventory."



import pytest
import torch

from idinn.demand import UniformDemand
from idinn.dual_controller import DualSourcingNeuralController
from idinn.sourcing_model import DualSourcingModel

torch.set_default_device('cpu')
# TODO: CUDA is numerically unstable,
#  may need new tests!


@pytest.fixture
def dual_sourcing_model_neural():
    return DualSourcingModel(
        regular_lead_time=2,
        expedited_lead_time=0,
        regular_order_cost=0,
        expedited_order_cost=20,
        holding_cost=5,
        shortage_cost=495,
        batch_size=256,
        init_inventory=6,
        demand_generator=UniformDemand(low=0, high=4),
    )


@pytest.fixture(scope="module")
def trained_neural_controller():
    dual_sourcing_model_train = DualSourcingModel(
        regular_lead_time=2,
        expedited_lead_time=0,
        regular_order_cost=0,
        expedited_order_cost=20,
        holding_cost=5,
        shortage_cost=495,
        batch_size=8,
        init_inventory=6,
        demand_generator=UniformDemand(low=0, high=4),
    )
    controller_neural = DualSourcingNeuralController(
        hidden_layers=[128, 64, 32, 16, 8, 4],
        activation=torch.nn.CELU(alpha=1),
    )
    controller_neural.fit(
        sourcing_model=dual_sourcing_model_train,
        sourcing_periods=100,
        validation_sourcing_periods=1,
        epochs=2,
        seed=1234,
    )
    return controller_neural


def test_neural_controller_init_layers(
    trained_neural_controller: DualSourcingNeuralController,
):
    assert (
        len(trained_neural_controller.model) == 14
    ), f"Model should have 14 layers, but got {len(trained_neural_controller.model)}"


def test_neural_controller_avg_cost(
    dual_sourcing_model_neural, trained_neural_controller
):
    # Calculate average cost
    avg_cost = trained_neural_controller.get_average_cost(
        dual_sourcing_model_neural, sourcing_periods=1000
    )
    assert (
        abs(avg_cost - 51133.42) < 1
    ), f"Average cost should be near 51133.42, but got {avg_cost}"


def test_neural_controller_simulate(
    dual_sourcing_model_neural, trained_neural_controller
):
    # TODO: Simulate is not working properly
    past_inventories, past_regular_orders, past_expedited_orders = (
        trained_neural_controller.simulate(
            dual_sourcing_model_neural, sourcing_periods=100
        )
    )

    # Validate simulation results
    assert (
        len(past_inventories) == 101
    ), "Simulation did not return correct number of inventory records."
    assert (
        len(past_regular_orders) == 101
    ), "Simulation did not return correct number of regular order records."
    assert (
        len(past_expedited_orders) == 101
    ), "Simulation did not return correct number of expedited order records."


def test_neural_controller_plot(dual_sourcing_model_neural, trained_neural_controller):
    # Ensure plotting does not raise errors
    trained_neural_controller.plot(dual_sourcing_model_neural, sourcing_periods=100)


def test_neural_controller_order_prediction(
    dual_sourcing_model_neural, trained_neural_controller
):
    # Predict orders using the trained controller
    regular_order, expedited_order = trained_neural_controller.predict(
        current_inventory=3, past_regular_orders=[1], past_expedited_orders=[0]
    )

    # Validate predictions
    assert (
        regular_order == 0
    ), "Predicted regular order should be 0, but got {regular_order}."
    assert (
        expedited_order == 0
    ), "Predicted expedited order should be 0, but got {expedited_order}."



import pytest
import torch

from idinn.demand import UniformDemand
from idinn.sourcing_model import SingleSourcingModel


@pytest.fixture
def single_sourcing_model():
    return SingleSourcingModel(
        lead_time=2,
        holding_cost=0.5,
        shortage_cost=1,
        init_inventory=10,
        demand_generator=UniformDemand(low=1, high=4),
        batch_size=3,
    )


def test_single_sourcing_model_initialization(
    single_sourcing_model: SingleSourcingModel,
):
    assert single_sourcing_model.get_lead_time() == 2, "Lead time should be 2."
    assert (
        single_sourcing_model.get_holding_cost() == 0.5
    ), "Holding cost should be 0.5."
    assert torch.all(
        torch.eq(single_sourcing_model.get_init_inventory(), torch.tensor([10.0]))
    ), "Inital inventory should be 10."


def test_single_sourcing_model_order(single_sourcing_model: SingleSourcingModel):
    q = torch.tensor([[5.0], [10.0], [15.0]])
    single_sourcing_model.order(q)
    past_orders = single_sourcing_model.get_past_orders()
    assert past_orders.shape == (
        3,
        2,
    ), f"Past orders tensor should have shape (3, 2), got {past_orders.shape}."
    assert torch.all(
        torch.eq(past_orders[:, -1], q.squeeze())
    ), f"Last order should match {q.squeeze()}."


def test_single_sourcing_model_update_inventories(
    single_sourcing_model: SingleSourcingModel,
):
    q = torch.tensor([[5.0], [10.0], [15.0]])
    single_sourcing_model.order(q, seed=42)
    past_inventories = single_sourcing_model.get_past_inventories()
    assert (
        past_inventories.shape == (3, 2)
    ), f"Past inventories tensor should have shape (3, 2), got {past_inventories.shape}."
    current_inventory = single_sourcing_model.get_current_inventory()
    assert (
        current_inventory.shape == (3, 1)
    ), f"Current inventory tensor should have shape (3, 1), got {current_inventory.shape}."


def test_single_sourcing_model_demands(single_sourcing_model: SingleSourcingModel):
    single_sourcing_model.order(torch.tensor([[10.0], [20.0], [30.0]]), seed=123)
    past_demands = single_sourcing_model.get_past_demands()
    assert past_demands.shape == (
        3,
        2,
    ), f"Past demands tensor should have shape (3, 2), got {past_demands.shape}."
    assert torch.all(past_demands[:, -1] >= 1) and torch.all(
        past_demands[:, -1] <= 4
    ), "Generated demands should be between 1 and 4."


def test_single_sourcing_model_reset(single_sourcing_model: SingleSourcingModel):
    single_sourcing_model.reset(batch_size=5)
    assert (
        single_sourcing_model.batch_size == 5
    ), "Batch size should be updated to 5 after reset."
    assert single_sourcing_model.get_past_orders().shape == (
        5,
        1,
    ), "Past orders tensor should have shape (5, 1)."
    assert single_sourcing_model.get_past_inventories().shape == (
        5,
        1,
    ), "Past inventories tensor should have shape (5, 1)."



from abc import ABCMeta, abstractmethod

import torch

class BaseDemand(metaclass=ABCMeta):
    @abstractmethod
    def sample(self, batch_size: int) -> torch.Tensor:
        """
        Generate demand for one period.

        Parameters
        ----------
        batch_size: int
            Size of generated demands which should correspond to the batch size or the number of SKUs.
        """
        pass

    def enumerate_support(self) -> dict:
        pass

    def get_min_demand(self) -> int:
        pass

    def get_max_demand(self) -> int:
        pass


class UniformDemand(BaseDemand):
    def __init__(self, low: int, high: int):
        self.distribution = torch.distributions.Uniform(low=low, high=high + 1)
        self.demand_prob = 1 / (high - low + 1)
        self.min_demand = low
        self.max_demand = high

    def sample(self, batch_size: int, batch_width: int = 1) -> torch.Tensor:
        return self.distribution.sample([batch_size, batch_width]).int()

    def enumerate_support(self) -> dict:
        return {
            x: 1 / (self.max_demand + 1 - self.min_demand)
            for x in range(self.min_demand, self.max_demand + 1)
        }

    def get_min_demand(self) -> int:
        return self.min_demand

    def get_max_demand(self) -> int:
        return self.max_demand


class CustomDemand(BaseDemand):
    def __init__(self, demand_prob: dict = None):
        from math import isclose

        # All demand values should be int
        for key in demand_prob:
            if not isinstance(key, int):
                raise TypeError(f"Demand values '{key}' is not an integer.")
        # All demand probabilities should be float
        for value in demand_prob.values():
            if not isinstance(value, (int, float)):
                raise TypeError(f"Demand probabilities '{value}' is not a float.")
        # Sum of probabilities should be close to 1
        total = sum(demand_prob.values())
        if not isclose(total, 1, abs_tol=1e-3):
            raise ValueError(
                f"The sum of demand probablities is {total}, which is not close to 1."
            )

        self.demand_prob = demand_prob

    def sample(self, batch_size: int, batch_width: int = 1) -> torch.Tensor:
        """
        Generate demand for one period.

        Parameters
        ----------
        batch_size: int
            Size of generated demands which should correspond to the batch size or the number of SKUs. If the size does not match the dimension of the elements from `demand_history`, demand will be upsampled or downsampled to match the size.
        """
        # Require Python >= 3.8 thus order is preserved (matches insertion order)
        # Draw dictionary keys with corresponding probabilities
        sampled_indices = torch.multinomial(
            torch.tensor(list(self.demand_prob.values())),
            num_samples=batch_size * batch_width,
            replacement=True,
        )
        return torch.tensor(list(self.demand_prob.keys()))[sampled_indices].reshape(
            batch_size, batch_width
        )

    def enumerate_support(self) -> dict:
        return self.demand_prob

    def get_min_demand(self) -> int:
        return min(self.demand_prob.keys())

    def get_max_demand(self) -> int:
        return max(self.demand_prob.keys())



from itertools import product
from typing import Dict as TypingDict, List as TypingList, Optional, Tuple, Union

import numpy as np
import torch
from numba import njit, types
from numba.typed import Dict, List

from ..sourcing_model import DualSourcingModel
from ..demand import UniformDemand
from .base import BaseDualController

import logging
from datetime import datetime

# Get root logger
logger = logging.getLogger()

class DynamicProgrammingController(BaseDualController):
    def __init__(self) -> None:
        self.sourcing_model= None
        self.qf = None
        self.vf = None
        logger.info("Initialized DynamicProgrammingController")

    @staticmethod
    @njit
    def __vf_update(
        demand_prob: TypingDict[int, float],
        min_demand: int,
        max_demand: int,
        ce: float,
        h: float,
        b: float,
        state: Tuple[int, ...],
        vf: TypingDict[Tuple[int, ...], float],
        actions: TypingList[Tuple[int, int]],
    ) -> Tuple[float, Optional[Tuple[int, int]]]:
        """
        vf_update is a function that calculates a single value iteration update.
        """
        best_action = None
        best_cost = 10e9

        for qe, qr in actions:
            # Immediate cost of action
            cost = qe * ce
            # Partial state update
            ip_e = state[0] + qe + state[1]
            pipeline = state[2:]

            for dem in range(int(min_demand), int(max_demand) + 1):
                ipe_new = int(ip_e) - dem
                this_state = (ipe_new,) + pipeline + (qr,)
                # If we jump to a state that is not in our list, we are not playing optimal,
                # so we can safely get out of here.
                if (this_state not in vf) or (vf[this_state] > 10e9 - 1.0):
                    cost = 10e9
                    break
                else:
                    # Careful: qr(t-1) has not arrived yet, we need to take it out
                    inv_on_hand = ipe_new - state[1]
                    inv_cost = inv_on_hand * h if inv_on_hand >= 0 else -inv_on_hand * b
                    cost += demand_prob[dem] * (inv_cost + vf[this_state])
            if cost < best_cost:
                best_cost = cost
                best_action = (qr, qe)
        return best_cost, best_action

    @staticmethod
    def __get_basestock_ub(
        exp_demand: float, lead_time: int, support: float, h: float, b: float
    ) -> float:
        """
        Get an upper bound on the single-source basestock level based on
        Hoeffding's inequality.
        """
        n = lead_time + 1
        base_stock_ub = n * exp_demand + support * np.sqrt(n * np.log(1 + b / h) / 2)
        return np.ceil(base_stock_ub)

    def fit(
        self,
        sourcing_model: DualSourcingModel,
        max_iterations: int = 1000000,
        tolerance: float = 10e-8,
        validation_freq: int = 100,
        log_freq: int = 100,
    ) -> None:
        """
        Fit the controller to the given sourcing model.

        Parameters
        ----------
        sourcing_model : DualSourcingModel
            The sourcing model to fit the controller to.
        max_iterations : int, default is 1000000
            Specifies the maximum number of iterations to run.
        tolerance : float, default is 10e-8
            Specifies the tolerance to check if the value function has converged.
        validation_freq : int, default is 100
            Specifies how many iteration to run before checking the tolerance is reached, e.g. `validation_freq=10` runs validation every 10 epochs.
        log_freq : int, default is 10
            Specifies how many training epochs to run before logging the training loss.
        """
        self.sourcing_model = sourcing_model

        # Check demand is uniform distributed
        if not isinstance(sourcing_model.demand_generator, UniformDemand):
            raise ValueError(
                "DynamicProgrammingController only supports uniform demand distribution."
            )
        # Check if the expedited_lead_time is 0
        if sourcing_model.expedited_lead_time != 0:
            raise ValueError(
                "DynamicProgrammingController only supports expedited_lead_time = 0."
            )
        
        start_time = datetime.now()
        logger.info(f"Starting dynamic programming at {start_time}")
        logger.info(f"Sourcing model parameters: batch_size={self.sourcing_model.batch_size}, "
                    f"lead_time={self.sourcing_model.lead_time}, init_inventory={self.sourcing_model.init_inventory.int().item()}, "
                    f"demand_generator={self.sourcing_model.demand_generator.__class__.__name__}")
        logger.info(f"Training parameters: max_iterations={max_iterations}, tolerance={tolerance}")

        min_demand = int(sourcing_model.demand_generator.get_min_demand())
        max_demand = int(sourcing_model.demand_generator.get_max_demand())
        exp_demand = (max_demand + min_demand) / 2.0
        support = max_demand - min_demand
        h = sourcing_model.get_holding_cost()
        b = sourcing_model.get_shortage_cost()
        ce = sourcing_model.get_expedited_order_cost()
        le = sourcing_model.get_expedited_lead_time()
        lr = sourcing_model.get_regular_lead_time()

        base_e = DynamicProgrammingController.__get_basestock_ub(
            exp_demand=exp_demand, lead_time=le, support=support, h=h, b=b
        )
        base_r = DynamicProgrammingController.__get_basestock_ub(
            exp_demand=exp_demand, lead_time=lr, support=support, h=h, b=b
        )
        min_ip = int(min(base_r, base_e) - max_demand)
        max_ip = int(max(base_r, base_e))
        max_order = max_ip + min_ip
        dim_pipeline = lr - le - 1

        demand_prob = Dict.empty(key_type=types.int64, value_type=types.float64)
        demand_prob_ = sourcing_model.demand_generator.enumerate_support()
        for k, v in demand_prob_.items():
            demand_prob[k] = v

        states_ = list(
            product(
                range(-min_ip, max_ip + 1),
                *(range(int(max_demand) + 1),) * int(dim_pipeline),
            )
        )
        states = List()
        for state in states_:
            states.append(state)

        actions_ = list(product(range(max_order), range(max_order)))
        actions = List()
        for action in actions_:
            actions.append(action)

        # Values can be initiated arbitrarily
        vals = np.repeat(1.0, len(states))
        vf_ = dict(zip(states, vals))
        vf = Dict.empty(
            key_type=types.UniTuple(types.int64, lr), value_type=types.float64
        )
        for k, v in vf_.items():
            vf[k] = v

        all_values = np.zeros(max_iterations, dtype=float)
        these_values = np.zeros(len(states))
        iteration_arr = []
        value_arr = []
        qf = {}
        val = 0

        for iteration in range(max_iterations):
            # We first store each newly updated state
            for idx, state in enumerate(states):
                these_values[idx] = DynamicProgrammingController.__vf_update(
                    demand_prob, min_demand, max_demand, ce, h, b, state, vf, actions
                )[0]
            # After the minimum for each state has been calculated, we update the states
            # If done in the same loop, converge sucks even more
            for idx, state in enumerate(states):
                vf[state] = these_values[idx]

            iter_vals = np.array([val for val in vf.values() if val < 10e8])
            this_average = np.mean(iter_vals)

            val = this_average / (iteration + 1)
            all_values[iteration] = val
            
            if iteration > 1 and iteration % log_freq == 0:
                logger.info(f"Epoch {iteration}/{max_iterations} - Value: {all_values[iteration]:.4f}")

            if iteration > 1 and iteration % validation_freq == 0:
                iteration_arr.append(iteration)
                value_arr.append(all_values[iteration])
                delta = all_values[iteration - 1] - all_values[iteration]
                if delta <= tolerance:
                    for state in states:
                        qa = DynamicProgrammingController.__vf_update(
                            demand_prob,
                            min_demand,
                            max_demand,
                            ce,
                            h,
                            b,
                            state,
                            vf,
                            actions,
                        )[1]
                        if qa is not None:
                            qf[state] = qa
                    break

        self.qf = qf
        self.vf = val

        end_time = datetime.now()
        duration = end_time - start_time
        logger.info(f"Dynamic programming completed at {end_time}")
        logger.info(f"Total training duration: {duration}")
        logger.info(f"Final best cost: {self.get_average_cost(self.sourcing_model, sourcing_periods=1000, seed=42):.4f}")

    def predict(
        self,
        current_inventory: Union[int, torch.Tensor],
        past_regular_orders: Optional[Union[TypingList[int], torch.Tensor]] = None,
        past_expedited_orders: Optional[Union[TypingList[int], torch.Tensor]] = None,
        output_tensor: bool = False,
    ) -> Union[Tuple[torch.Tensor, torch.Tensor], Tuple[int, int]]:
        """

        Parameters
        ----------
        current_inventory : int, or torch.Tensor
            Current inventory.
        past_regular_orders : list, or torch.Tensor, optional
            Past regular orders. If the length of `past_regular_orders` is lower than `regular_lead_time`, it will be padded with zeros. If the length of `past_regular_orders` is higher than `regular_lead_time`, only the last `regular_lead_time` orders will be used during inference.
        past_expedited_orders : list, or torch.Tensor, optional
            Past expedited orders. Since `expedited_lead_time` is assumed to be 0 for DynamicProgrammingController and the batch size is assumed to be 1, the input of `past_expedited_orders` is optional and will be ignored.
        output_tensor : bool, default is False
            If True, the replenishment order quantity will be returned as a torch.Tensor. Otherwise, it will be returned as an integer.

        """
        if self.sourcing_model is None:
            raise AttributeError("The controller is not trained.")

        regular_lead_time = self.sourcing_model.get_regular_lead_time()

        current_inventory = self._check_current_inventory(current_inventory)
        past_regular_orders = self._check_past_orders(
            past_regular_orders, regular_lead_time
        )

        first = (
            current_inventory.squeeze()
            + past_regular_orders.squeeze()[-regular_lead_time]
        )
        second = past_regular_orders.squeeze()[-regular_lead_time + 1 :]
        key = tuple([int(first)] + second.int().tolist())
        if output_tensor:
            return torch.tensor([[self.qf[key][0]]]), torch.tensor([[self.qf[key][1]]])
        else:
            return self.qf[key]

    def reset(self) -> None:
        self.qf = None
        self.vf = None
        self.sourcing_model = None



import pytest
import torch

from idinn.demand import UniformDemand
from idinn.single_controller import SingleSourcingNeuralController
from idinn.sourcing_model import SingleSourcingModel


@pytest.fixture
def single_sourcing_model_neural():
    return SingleSourcingModel(
        lead_time=0,
        holding_cost=5,
        shortage_cost=495,
        batch_size=32,
        init_inventory=10,
        demand_generator=UniformDemand(low=1, high=4),
    )


@pytest.fixture(scope="module")
def trained_neural_controller():
    sourcing_model_train = SingleSourcingModel(
        lead_time=0,
        holding_cost=5,
        shortage_cost=495,
        batch_size=4,
        init_inventory=10,
        demand_generator=UniformDemand(low=1, high=4),
    )
    neural_controller = SingleSourcingNeuralController(
        hidden_layers=[2], activation=torch.nn.CELU(alpha=1)
    )
    neural_controller.fit(
        sourcing_model=sourcing_model_train,
        sourcing_periods=5,
        validation_sourcing_periods=1,
        epochs=2,
        seed=1,
    )
    return neural_controller


def test_neural_controller_init_layers(
    trained_neural_controller: SingleSourcingNeuralController,
):
    assert (
        len(trained_neural_controller.model) == 4
    ), f"Model should have 4 layers, but got {len(trained_neural_controller.model)}"


def test_neural_controller_avg_cost(
    single_sourcing_model_neural, trained_neural_controller
):
    # Calculate average cost
    avg_cost = trained_neural_controller.get_average_cost(
        single_sourcing_model_neural, sourcing_periods=1000, seed=42
    )
    assert (
        abs(avg_cost - 4681.83) < 1
    ), f"Average cost should be near 4681.83 but got {avg_cost}"


def test_neural_controller_simulate(
    single_sourcing_model_neural, trained_neural_controller
):
    # Simulate sourcing
    past_inventories, past_orders = trained_neural_controller.simulate(
        single_sourcing_model_neural, sourcing_periods=100
    )

    # Validate simulation results
    assert (
        len(past_inventories) == 101
    ), "Simulation did not return correct number of inventory records."
    assert (
        len(past_orders) == 101
    ), "Simulation did not return correct number of order records."


def test_neural_controller_plot(
    single_sourcing_model_neural, trained_neural_controller
):
    # Ensure plotting does not raise errors
    trained_neural_controller.plot(single_sourcing_model_neural, sourcing_periods=100)


def test_neural_controller_predict(trained_neural_controller):
    # Predict order quantity
    predicted_order = trained_neural_controller.predict(current_inventory=0)

    # Validate prediction
    assert predicted_order is not None, "Predicted order should not be None."
    assert int(predicted_order) == 0, "Predicted order quantity should be 4."



from typing import Union

import torch

from .demand import CustomDemand, UniformDemand


class BaseSourcingModel:
    def __init__(
        self,
        holding_cost: float,
        shortage_cost: float,
        init_inventory: float,
        demand_generator: Union[UniformDemand, CustomDemand],
        batch_size: int,
        lead_time: int = None,
        regular_lead_time: int = None,
        expedited_lead_time: int = None,
        regular_order_cost: float = None,
        expedited_order_cost: float = None,
    ):
        self.holding_cost = holding_cost
        self.shortage_cost = shortage_cost
        self.init_inventory = torch.tensor(
            [init_inventory], requires_grad=True, dtype=torch.float
        )
        self.batch_size = batch_size
        self.lead_time = lead_time
        self.regular_lead_time = regular_lead_time
        self.expedited_lead_time = expedited_lead_time
        self.regular_order_cost = regular_order_cost
        self.expedited_order_cost = expedited_order_cost
        self.demand_generator = demand_generator
        self.reset()

    def reset(self, batch_size: int = None) -> None:
        if batch_size is not None and self.batch_size != batch_size:
            self.batch_size = batch_size

        self.past_inventories = self.get_init_inventory().repeat(self.batch_size, 1)
        self.past_demands = torch.zeros(self.batch_size, 1)

        if self.lead_time is not None:
            self.past_orders = torch.zeros(self.batch_size, 1)

        elif (
            self.regular_lead_time is not None and self.expedited_lead_time is not None
        ):
            if self.regular_lead_time < self.expedited_lead_time:
                raise ValueError(
                    "`regular_lead_time` must be greater than or equal to expedited_lead_time."
                )
            self.past_regular_orders = torch.zeros(self.batch_size, 1)
            self.past_expedited_orders = torch.zeros(self.batch_size, 1)
        else:
            raise ValueError(
                "Either `lead_time` or (`regular_lead_time` and `expedited_lead_time`) must be provided."
            )

    def get_holding_cost(self) -> float:
        return self.holding_cost

    def get_shortage_cost(self) -> float:
        return self.shortage_cost

    def get_init_inventory(self) -> torch.Tensor:
        init_inventory = (
            self.init_inventory - torch.frac(self.init_inventory).clone().detach()
        )
        return init_inventory

    def get_past_inventories(self) -> torch.Tensor:
        return self.past_inventories

    def get_past_demands(self) -> torch.Tensor:
        return self.past_demands

    def get_current_inventory(self) -> torch.Tensor:
        # Keep dim when slicing past_inventories
        # https://stackoverflow.com/questions/57237352/why-does-torch-slice-lose-dimension
        # https://discuss.pytorch.org/t/solved-simple-question-about-keep-dim-when-slicing-the-tensor/9280
        return self.past_inventories[:, [-1]]


class SingleSourcingModel(BaseSourcingModel):
    def __init__(
        self,
        lead_time: int,
        holding_cost: float,
        shortage_cost: float,
        init_inventory: float,
        demand_generator: Union[UniformDemand, CustomDemand],
        batch_size: int = 1,
    ):
        """
        Parameters
        ----------
        lead_time : int
            The lead time for orders.
        holding_cost : float
            The cost of holding inventory.
        shortage_cost : float
            The cost of inventory shortage.
        init_inventory : float
            The initial inventory.
        demand_generator : Demand
        batch_size : int, optional
            The batch size for orders. default is 1.
        """
        super().__init__(
            lead_time=lead_time,
            holding_cost=holding_cost,
            shortage_cost=shortage_cost,
            init_inventory=init_inventory,
            batch_size=batch_size,
            demand_generator=demand_generator,
        )

    def get_lead_time(self) -> int:
        return self.lead_time

    def get_past_orders(self) -> torch.Tensor:
        return self.past_orders

    def order(self, q: torch.Tensor, seed: int = None) -> None:
        """
        Orders items to the inventory and update the inventory with generated demands.

        Parameters
        ----------
        q : torch.Tensor
            The quantity of items to order.
        seed : int, optional
            Random seed for reproducibility.
        """
        if seed is not None:
            torch.manual_seed(seed)
        # Current orders are added to past_orders
        self.past_orders = torch.cat([self.past_orders, q], dim=1)
        # Past orders arrived, if past orders are not available, then arrived order is 0
        if self.past_orders.shape[1] >= 1 + self.lead_time:
            arrived_order = self.past_orders[:, [-1 - self.lead_time]]
        else:
            arrived_order = torch.zeros(self.batch_size, 1)
        # Generate current demand
        current_demand = self.demand_generator.sample(self.batch_size)
        # Current demand are added to past_demands
        self.past_demands = torch.cat([self.past_demands, current_demand], dim=1)
        # Update inventory
        current_inventory = (
            self.get_current_inventory() + arrived_order - current_demand
        )
        # Current inventories are added to past_inventories
        self.past_inventories = torch.cat(
            [self.past_inventories, current_inventory], dim=1
        )


class DualSourcingModel(BaseSourcingModel):
    def __init__(
        self,
        regular_lead_time: int,
        expedited_lead_time: int,
        regular_order_cost: float,
        expedited_order_cost: float,
        holding_cost: float,
        shortage_cost: float,
        init_inventory: float,
        demand_generator: Union[UniformDemand, CustomDemand],
        batch_size: int = 1,
    ):
        """
        Parameters
        ----------
        regular_lead_time : int
            The lead time for regular orders.
        expedited_lead_time : int
            The lead time for expedited orders.
        regular_order_cost : float
            The cost of placing a regular order.
        expedited_order_cost : float
            The cost of placing an expedited order.
        holding_cost : float
            The cost of holding inventory.
        shortage_cost : float
            The cost of shortage.
        init_inventory : float
            The initial inventory.
        batch_size : int, default is 1
            The batch size for orders.
        demand : Iterable, torch.Tensor, np.array, or pd.Series, optional
            The array for outputting values of demands.
        demand_distribution : str, default is `uniform`.
            Distribution for generated demand when `demand` is not specified.
        demand_low : int, default is 1.
            Lower bound for generated demand when `demand` is not specified.
        demand_high : int, default is 4.
            Higher bound for generated demand when `demand` is not specified.
        """
        super().__init__(
            regular_lead_time=regular_lead_time,
            expedited_lead_time=expedited_lead_time,
            regular_order_cost=regular_order_cost,
            expedited_order_cost=expedited_order_cost,
            holding_cost=holding_cost,
            shortage_cost=shortage_cost,
            init_inventory=init_inventory,
            batch_size=batch_size,
            demand_generator=demand_generator,
        )

    def get_past_regular_orders(self) -> torch.Tensor:
        return self.past_regular_orders

    def get_past_expedited_orders(self) -> torch.Tensor:
        return self.past_expedited_orders

    def get_last_regular_order(self) -> torch.Tensor:
        return self.past_regular_orders[:, [-1]]

    def get_last_expedited_order(self) -> torch.Tensor:
        return self.past_expedited_orders[:, [-1]]

    def get_regular_lead_time(self) -> int:
        return self.regular_lead_time

    def get_expedited_lead_time(self) -> int:
        return self.expedited_lead_time

    def get_regular_order_cost(self) -> float:
        return self.regular_order_cost

    def get_expedited_order_cost(self) -> float:
        return self.expedited_order_cost

    def order(
        self, regular_q: torch.Tensor, expedited_q: torch.Tensor, seed: int = None
    ) -> None:
        """
        Orders items to the inventory and update the inventory with generated demands.

        Parameters
        ----------
        regular_q : torch.Tensor
            The quantity of items to order from the regular supplier.
        expedited_q : torch.Tensor
            The quantity of items to order from the expedited supplier.
        seed : int, optional
            Random seed for reproducibility.
        """
        if seed is not None:
            torch.manual_seed(seed)
        if not isinstance(regular_q, torch.Tensor):
            regular_q = torch.tensor([[regular_q]])
        if not isinstance(expedited_q, torch.Tensor):
            expedited_q = torch.tensor([[expedited_q]])
        # Current regular order are added to past_regular_orders
        self.past_regular_orders = torch.cat(
            [self.past_regular_orders, regular_q], dim=1
        )
        # Current expedited order are added to past_expedited_orders
        self.past_expedited_orders = torch.cat(
            [self.past_expedited_orders, expedited_q], dim=1
        )
        # Past regular orders arrived,
        # if past regular orders are not available, then arrived order is 0
        if self.past_regular_orders.shape[1] >= 1 + self.regular_lead_time:
            arrived_regular_orders = self.past_regular_orders[
                :, [-1 - self.regular_lead_time]
            ]
        else:
            arrived_regular_orders = torch.zeros(self.batch_size, 1)
        # Past expedited orders arrived,
        # if past expedited orders are not available, then arrived order is 0
        if self.past_expedited_orders.shape[1] >= 1 + self.expedited_lead_time:
            arrived_expedited_orders = self.past_expedited_orders[
                :, [-1 - self.expedited_lead_time]
            ]
        else:
            arrived_expedited_orders = torch.zeros(self.batch_size, 1)
        # Generate current demand
        current_demand = self.demand_generator.sample(self.batch_size)
        # Current demand are added to past_demands
        self.past_demands = torch.cat([self.past_demands, current_demand], dim=1)
        # Update inventory
        current_inventory = (
            self.get_current_inventory()
            + arrived_expedited_orders
            + arrived_regular_orders
            - current_demand
        )
        self.past_inventories = torch.cat(
            [self.past_inventories, current_inventory], dim=1
        )



from typing import List, Optional, Union

import numpy as np
import torch
from torch.utils.tensorboard import SummaryWriter

from ..sourcing_model import SingleSourcingModel
from .base import BaseSingleController

import logging
from datetime import datetime

# Add logger setup at class level
logger = logging.getLogger(__name__)

class SingleSourcingNeuralController(torch.nn.Module, BaseSingleController):
    """
    SingleSourcingNeuralController is a neural network-based controller for inventory optimization in a single-sourcing scenario.

    Parameters
    ----------
    hidden_layers : list, default is [2]
        List of integers representing the number of units in each hidden layer.
    activation : torch.nn.Module, default is torch.nn.CELU(alpha=1)
        Activation function to be used in the hidden layers.

    Attributes
    ----------
    hidden_layers : list
        List of integers representing the number of units in each hidden layer.
    activation : torch.nn.Module
        Activation function used in the hidden layers.
    architecture : torch.nn.Sequential
        Sequential stack of linear layers and activation functions.

    Methods
    -------
    init_layers(lead_time)
        Initialize the layers of the neural network based on the lead time.
    forward(current_inventory, past_orders)
        Perform forward pass through the neural network.
    get_total_cost(sourcing_model, sourcing_periods, seed=None)
        Calculate the total cost over a given number of sourcing periods.
    train(sourcing_model, sourcing_periods, epochs, ...)
        Train the neural network controller using the sourcing model and specified parameters.
    simulate(sourcing_model, sourcing_periods)
        Simulate the inventory and order quantities over a given number of sourcing periods.
    plot(sourcing_model, sourcing_periods)
        Plot the inventory and order quantities over a given number of sourcing periods.
    """

    def __init__(self, hidden_layers: List[int] = [2], activation: torch.nn.Module = torch.nn.CELU(alpha=1)):
        super().__init__()
        self.sourcing_model = None
        self.hidden_layers = hidden_layers
        self.activation = activation
        self.model = None
        logger.info(f"Initialized SingleSourcingNeuralController with hidden_layers={hidden_layers}")

    def init_layers(self) -> None:
        """
        Initialize the layers of the neural network based on the lead time.

        Returns
        -------
        None
        """
        lead_time = self.sourcing_model.get_lead_time()
        architecture = [
            torch.nn.Linear(lead_time + 1, self.hidden_layers[0]),
            self.activation,
        ]
        for i in range(len(self.hidden_layers)):
            if i < len(self.hidden_layers) - 1:
                architecture += [
                    torch.nn.Linear(self.hidden_layers[i], self.hidden_layers[i + 1]),
                    self.activation,
                ]
        architecture += [
            torch.nn.Linear(self.hidden_layers[-1], 1, bias=False),
            torch.nn.ReLU(),
        ]
        self.model = torch.nn.Sequential(*architecture)
    
    def prepare_inputs(
        self,
        current_inventory: torch.Tensor,
        past_orders: torch.Tensor
    ) -> torch.Tensor:
        lead_time = self.sourcing_model.get_lead_time()
        current_inventory = self._current_inventory_check(current_inventory)
        past_orders = self._past_orders_check(past_orders, lead_time)
        
        if lead_time == 0:
            inputs = current_inventory
        elif lead_time > 0:
            inputs = torch.cat([current_inventory, past_orders[:, -lead_time:]], dim=1)
        else:
            raise ValueError("`lead_time` cannot be less than 0")
        
        return inputs

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        h = self.model(inputs)
        q = h - torch.frac(h).clone().detach()
        return q

    def predict(
        self,
        current_inventory: Union[int, torch.Tensor],
        past_orders: Optional[Union[List[int], torch.Tensor]] = None,
        output_tensor: bool = False,
    ) -> Union[torch.Tensor, int]:
        """
        Perform forward pass through the neural network.

        Parameters
        ----------
        current_inventory : int, or torch.Tensor
            Current inventory levels.
        past_orders : list, or torch.Tensor, optional
            Past order quantities. If the length of `past_orders` is lower than `lead_time`, it will be padded with zeros. If the length of `past_orders` is higher than `lead_time`, only the last `lead_time` orders will be used during inference.
        output_tensor : bool, default is False
            If True, the replenishment order quantity will be returned as a torch.Tensor. Otherwise, it will be returned as an integer.

        Returns
        -------
        torch.Tensor
            Order quanty calculated by the neural network.
        """
        if self.sourcing_model is None:
            raise AttributeError("The controller is not trained.")

        
        inputs = self.prepare_inputs(current_inventory, past_orders)
        q = self.forward(inputs)

        if output_tensor:
            return q
        else:
            return q.int().item()

    def fit(
        self,
        sourcing_model: SingleSourcingModel,
        sourcing_periods: int,
        epochs: int,
        validation_sourcing_periods: Optional[int] = None,
        validation_freq: int = 50,
        log_freq: int = 50,
        init_inventory_lr: float = 1e-1,
        init_inventory_freq: int = 4,
        parameters_lr: float = 3e-3,
        tensorboard_writer: Optional[SummaryWriter] = None,
        seed: Optional[int] = None,
    ) -> None:
        """
        Train the neural network controller using the sourcing model and specified parameters.

        Parameters
        ----------
        sourcing_model : SourcingModel
            The sourcing model for training.
        sourcing_periods : int
            The number of sourcing periods for training.
        epochs : int
            The number of training epochs.
        validation_sourcing_periods : int, optional
            The number of sourcing periods for validation.
        validation_freq : int, default is 10
            Only relevant if `validation_sourcing_periods` is provided. Specifies how many training epochs to run before a new validation run is performed, e.g. `validation_freq=10` runs validation every 10 epochs.
        log_freq : int, default is 10
            Specifies how many training epochs to run before logging the training loss.
        init_inventory_freq : int, default is 4
            Specifies how many parameter updating epochs to run before initial inventory is updated. e.g. `init_inventory_freq=4` updates initial inventory after updating parameters for 4 epochs.
        init_inventory_lr : float, default is 1e-1
            Learning rate for initial inventory.
        parameters_lr : float, default is 3e-3
            Learning rate for updating neural network parameters.
        tensorboard_writer : tensorboard.SummaryWriter, optional
            Tensorboard writer for logging.
        seed : int, optional
            Random seed for reproducibility.
        """
        # Store sourcing model in self.sourcing_model
        self.sourcing_model = sourcing_model

        if seed is not None:
            torch.manual_seed(seed)

        if self.model is None:
            self.init_layers()

        start_time = datetime.now()
        logger.info(f"Starting single sourcing neural network training at {start_time}")
        logger.info(f"Sourcing model parameters: batch_size={self.sourcing_model.batch_size}, "
                    f"lead_time={self.sourcing_model.lead_time}, init_inventory={self.sourcing_model.init_inventory.int().item()}, "
                    f"demand_generator={self.sourcing_model.demand_generator.__class__.__name__}")
        logger.info(f"Training parameters: epochs={epochs}, sourcing_periods={sourcing_periods}, "
                    f"validation_periods={validation_sourcing_periods}, learning_rate={parameters_lr}")

        optimizer_init_inventory = torch.optim.RMSprop(
            [sourcing_model.init_inventory], lr=init_inventory_lr
        )
        optimizer_parameters = torch.optim.RMSprop(self.parameters(), lr=parameters_lr)
        min_cost = np.inf

        for epoch in range(epochs):
            # Clear grad cache
            optimizer_parameters.zero_grad()
            optimizer_init_inventory.zero_grad()
            # Reset the sourcing model with the learned init inventory
            sourcing_model.reset()
            logger.debug(f"Reset the sourcing model with the learned init inventory at epoch {epoch}")
            total_cost = self.get_total_cost(sourcing_model, sourcing_periods)
            total_cost.backward()
            # Gradient descend
            if epoch % init_inventory_freq == 0:
                optimizer_init_inventory.step()
            else:
                optimizer_parameters.step()
            # Save the best model
            if validation_sourcing_periods is not None and epoch % 10 == 0:
                eval_cost = self.get_total_cost(
                    sourcing_model, validation_sourcing_periods
                )
                if eval_cost < min_cost:
                    min_cost = eval_cost
                    best_state = self.model.state_dict()
            else:
                if total_cost < min_cost:
                    min_cost = total_cost
                    best_state = self.model.state_dict()
            # Log train loss
            if tensorboard_writer is not None:
                
                tensorboard_writer.add_scalar(
                    "Avg. cost per period/train", total_cost / sourcing_periods, epoch
                )
                if (
                    validation_sourcing_periods is not None
                    and epoch % validation_freq == 0
                ):
                    # Log validation loss
                    eval_cost = self.get_total_cost(
                        sourcing_model, validation_sourcing_periods
                    )
                    tensorboard_writer.add_scalar(
                        "Avg. cost per period/val",
                        eval_cost / validation_sourcing_periods,
                        epoch,
                    )
                logger.debug(f"Wrote to tensorboard at epoch {epoch}")
                tensorboard_writer.flush()

            end_time = datetime.now()
            duration = end_time - start_time
            per_epoch_time = duration.total_seconds() / (epoch + 1)  # seconds per epoch
            remaining_time = (epochs - epoch) * per_epoch_time

            if epoch % log_freq == 0:
                logger.info(f"Epoch {epoch}/{epochs}"
                            f" - Training cost: {total_cost/sourcing_periods:.4f}"
                            f" - Per epoch time: {per_epoch_time:.2f} seconds"
                            f" - Est. Remaining time: {int(remaining_time)} seconds."
                       )
                
            if validation_sourcing_periods is not None and epoch % validation_freq == 0:
                logger.info(f"Epoch {epoch}/{epochs}"
                            f" - Validation cost: {eval_cost/validation_sourcing_periods:.4f}"
                            f" - Per epoch time: {per_epoch_time:.2f} seconds"
                            f" - Est. Remaining time: {int(remaining_time)} seconds."
                            )

        # Load the best model
        self.model.load_state_dict(best_state)

        end_time = datetime.now()
        duration = end_time - start_time
        logger.info(f"Training completed at {end_time}")
        logger.info(f"Total training duration: {duration}")
        logger.info(f"Final best cost: {min_cost/sourcing_periods:.4f}")

    def reset(self) -> None:
        self.sourcing_model = None
        self.model = None

    def save(self, path: str) -> None:
        torch.save(self.model, path)

    def load(self, path: str) -> None:
        self.model = torch.load(path, weights_only=False)



from abc import ABCMeta, abstractmethod

import torch


class BaseSingleController(metaclass=ABCMeta):
    @abstractmethod
    def fit(self, sourcing_model):
        """
        Fit the controller to the sourcing model.
        """
        pass

    @abstractmethod
    def predict(self, current_inventory, past_orders, output_tensor):
        """
        Predict the replenishment order quantity.
        """
        pass

    @abstractmethod
    def reset(self):
        """
        Reset the controller to the initial state.
        """
        pass

    def _current_inventory_check(self, current_inventory):
        """
        Check and convert types of `current_inventory` for `predict()`.
        """
        if isinstance(current_inventory, int):
            current_inventory = torch.tensor([[current_inventory]], dtype=torch.float32)
        elif isinstance(current_inventory, torch.Tensor):
            pass
        else:
            raise TypeError("`current_inventory`'s type is not supported.")

        return current_inventory

    def _past_orders_check(self, past_orders, lead_time):
        """
        Check and convert types of `past_orders` for `predict()`. Pad `past_orders` with zeros if it is too short.
        """
        if past_orders is None:
            past_orders = torch.zeros(1, lead_time)
        elif isinstance(past_orders, list):
            past_orders = torch.tensor([past_orders], dtype=torch.float32)
        elif isinstance(past_orders, torch.Tensor):
            pass
        else:
            raise TypeError("`past_orders`'s type is not supported.")

        order_len = past_orders.shape[1]
        if order_len < lead_time:
            return torch.nn.functional.pad(past_orders, (lead_time - order_len, 0))
        else:
            return past_orders

    def get_last_cost(self, sourcing_model):
        """
        Calculate the cost for the latest period of the sourcing model.

        Parameters
        ----------
        sourcing_model : SourcingModel
            The sourcing model.

        Returns
        -------
        float
            The last cost.

        """
        shortage_cost = sourcing_model.get_shortage_cost()
        holding_cost = sourcing_model.get_holding_cost()
        current_inventory = sourcing_model.get_current_inventory()
        last_cost = holding_cost * torch.relu(
            current_inventory
        ) + shortage_cost * torch.relu(-current_inventory)
        return last_cost

    def get_total_cost(self, sourcing_model, sourcing_periods, seed=None):
        """
        Calculate the total cost for single-sourcing optimization.

        Parameters
        ----------
        sourcing_model : SourcingModel
            The sourcing model.
        sourcing_periods : int
            Number of sourcing periods.
        seed : int, optional
            Random seed for reproducibility.

        Returns
        -------
        float
            The total cost.
        """
        if seed is not None:
            torch.manual_seed(seed)

        total_cost = 0
        for i in range(sourcing_periods):
            current_inventory = sourcing_model.get_current_inventory()
            past_orders = sourcing_model.get_past_orders()
            q_t = self.predict(current_inventory, past_orders, output_tensor=True)
            sourcing_model.order(q_t)
            last_cost = self.get_last_cost(sourcing_model)
            total_cost += last_cost.mean()
        return total_cost

    def get_average_cost(self, sourcing_model, sourcing_periods, seed=None):
        """
        Calculate the average cost for single-sourcing optimization.

        Parameters
        ----------
        sourcing_model : SourcingModel
            The sourcing model.
        sourcing_periods : int
            Number of sourcing periods.
        seed : int, optional
            Random seed for reproducibility.

        Returns
        -------
        float
            The average cost.
        """
        return (
            self.get_total_cost(sourcing_model, sourcing_periods, seed)
            / sourcing_periods
        )

    def simulate(self, sourcing_model, sourcing_periods, seed=None):
        """
        Simulate the sourcing model's output using the given controller.

        Parameters
        ----------
        sourcing_model : SingleSourcingModel
            The sourcing model.
        sourcing_periods : int
            Number of sourcing periods.
        seed : int, optional
            Random seed for reproducibility.

        Returns
        -------
        past_inventories : list
            List of past inventories.
        past_orders : list
            List of past orders.

        """
        if seed is not None:
            torch.manual_seed(seed)
        sourcing_model.reset(batch_size=1)
        for i in range(sourcing_periods):
            current_inventory = sourcing_model.get_current_inventory()
            past_orders = sourcing_model.get_past_orders()
            q = self.predict(current_inventory, past_orders, output_tensor=True)
            sourcing_model.order(q)
        past_inventories = sourcing_model.get_past_inventories()[0, :].detach().numpy()
        past_orders = sourcing_model.get_past_orders()[0, :].detach().numpy()
        return past_inventories, past_orders

    def plot(self, sourcing_model, sourcing_periods, linewidth=1, seed=None):
        """
        Plot the inventory and order quantities over a given number of sourcing periods.

        Parameters
        ----------
        sourcing_model : SingleSourcingModel
            The sourcing model to be used for plotting.
        sourcing_periods : int
            The number of sourcing periods for plotting.
        linewidth : int, default is 1
            The width of the line in the step plots.
        seed : int, optional
            Random seed for reproducibility.
        """
        from matplotlib import pyplot as plt

        past_inventories, past_orders = self.simulate(
            sourcing_model=sourcing_model, sourcing_periods=sourcing_periods, seed=seed
        )
        fig, ax = plt.subplots(ncols=2, figsize=(10, 4))

        ax[0].step(
            range(sourcing_periods),
            past_inventories[-sourcing_periods:],
            linewidth=linewidth,
            color="tab:blue",
        )
        ax[0].yaxis.get_major_locator().set_params(integer=True)
        ax[0].set_title("Inventory")
        ax[0].set_xlabel("Period")
        ax[0].set_ylabel("Quantity")

        ax[1].step(
            range(sourcing_periods),
            past_orders[-sourcing_periods:],
            linewidth=linewidth,
            color="tab:orange",
        )
        ax[1].yaxis.get_major_locator().set_params(integer=True)
        ax[1].set_title("Order")
        ax[1].set_xlabel("Period")
        ax[1].set_ylabel("Quantity")

        return fig, ax



import pandas as pd
from tensorboard.backend.event_processing.event_accumulator import EventAccumulator


def tflog2pandas(path):
    """
    Convert TensorFlow log files to pandas DataFrame.

    Parameters
    ----------
    path : str
        Path to the TensorFlow log directory.

    Returns
    -------
    pd.DataFrame
        DataFrame containing the scalar metrics.
    """

    try:
        event_acc = EventAccumulator(path)
        event_acc.Reload()
        tags = event_acc.Tags()["scalars"]
        data_ = []
        for tag in tags:
            event_list = event_acc.Scalars(tag)
            values = [x.value for x in event_list]
            step = [x.step for x in event_list]
            data = pd.DataFrame(
                {
                    "metric": [tag.replace("Avg. cost per period/", "")] * len(step),
                    "value": values,
                    "step": step,
                }
            )
            data_.append(data)
        data = pd.concat(data_)
        data = data.loc[data["step"] >= 100]
    except Exception:
        print(f"Event file possibly corrupt: {path}")
        return pd.DataFrame()
    return data
